/*
 * pg_seclabel.c
 *
 * Portions Copyright (c) 1996-2010, PostgreSQL Global Development Group
 * Portions Copyright (c) 1994, Regents of the University of California
 */
#include "postgres.h"

#include "access/genam.h"
#include "access/heapam.h"
#include "access/sysattr.h"
#include "access/tupdesc.h"
#include "catalog/catalog.h"
#include "catalog/indexing.h"
#include "catalog/pg_aggregate.h"
#include "catalog/pg_amop.h"
#include "catalog/pg_amproc.h"
#include "catalog/pg_attrdef.h"
#include "catalog/pg_auth_members.h"
#include "catalog/pg_constraint.h"
#include "catalog/pg_db_role_setting.h"
#include "catalog/pg_depend.h"
#include "catalog/pg_description.h"
#include "catalog/pg_enum.h"
#include "catalog/pg_index.h"
#include "catalog/pg_inherits.h"
#include "catalog/pg_largeobject.h"
#include "catalog/pg_namespace.h"
#include "catalog/pg_rewrite.h"
#include "catalog/pg_seclabel.h"
#include "catalog/pg_shdepend.h"
#include "catalog/pg_shdescription.h"
#include "catalog/pg_trigger.h"
#include "catalog/pg_ts_config_map.h"
#include "executor/spi.h"
#include "miscadmin.h"
#include "sepgsql/sepgsql.h"
#include "sepgsql/hooks.h"
#include "utils/builtins.h"
#include "utils/fmgroids.h"
#include "utils/lsyscache.h"
#include "utils/rel.h"
#include "utils/syscache.h"
#include "utils/tqual.h"

bool ignore_security_label_input;

void
seclabelOnCreateDatabase(Oid src_datid, Oid dst_datid)
{
	Relation		rel;
	ScanKeyData		keys[1];
	SysScanDesc		scan;
	HeapTuple		oldtup, newtup;
	Datum			values[Natts_pg_seclabel];
	bool			nulls[Natts_pg_seclabel];
	bool			replaces[Natts_pg_seclabel];

	/* Scan all entries with pg_seclabel.datid = src_datid */
	ScanKeyInit(&keys[0],
				Anum_pg_seclabel_datid,
				BTEqualStrategyNumber, F_OIDEQ,
				ObjectIdGetDatum(src_datid));

	rel = heap_open(SecLabelRelationId, RowExclusiveLock);

	scan = systable_beginscan(rel, SecLabelSecidIndexId, true,
							  SnapshotNow, 1, keys);

	/* corresponding entries will be inserted with new datid */
	memset(values, 0, sizeof(values));
	memset(nulls, false, sizeof(nulls));
	memset(replaces, false, sizeof(replaces));

	values[Anum_pg_seclabel_datid - 1] = ObjectIdGetDatum(dst_datid);
	replaces[Anum_pg_seclabel_datid - 1] = true;
	while (HeapTupleIsValid(oldtup = systable_getnext(scan)))
	{
		newtup = heap_modify_tuple(oldtup, RelationGetDescr(rel),
								   values, nulls, replaces);
		simple_heap_insert(rel, newtup);

		CatalogUpdateIndexes(rel, newtup);

		heap_freetuple(newtup);
	}
	systable_endscan(scan);

	heap_close(rel, RowExclusiveLock);
}

void
seclabelOnDropDatabase(Oid datid)
{
	Relation		rel;
	ScanKeyData		keys[1];
	SysScanDesc		scan;
	HeapTuple		tuple;

	/* Scan all entries with pg_seclabel.datid = datid */
	ScanKeyInit(&keys[0],
				Anum_pg_seclabel_datid,
				BTEqualStrategyNumber, F_OIDEQ,
				ObjectIdGetDatum(datid));

	rel = heap_open(SecLabelRelationId, RowExclusiveLock);

	scan = systable_beginscan(rel, SecLabelSecidIndexId, true,
							  SnapshotNow, 1, keys);

	while (HeapTupleIsValid(tuple = systable_getnext(scan)))
	{
		simple_heap_delete(rel, &tuple->t_self);
	}

	systable_endscan(scan);

	heap_close(rel, RowExclusiveLock);
}

void
seclabelOnDropTable(Oid relid)
{
	Relation		rel;
	SysScanDesc		scan;
	ScanKeyData		key[2];
	HeapTuple		tuple;
	Oid				database_oid;

	database_oid = (IsSharedRelation(relid) ? InvalidOid : MyDatabaseId);
	ScanKeyInit(&key[0],
				Anum_pg_seclabel_datid,
				BTEqualStrategyNumber, F_OIDEQ,
				ObjectIdGetDatum(database_oid));
	ScanKeyInit(&key[1],
				Anum_pg_seclabel_relid,
				BTEqualStrategyNumber, F_OIDEQ,
				ObjectIdGetDatum(relid));

	rel = heap_open(SecLabelRelationId, RowExclusiveLock);
	scan = systable_beginscan(rel, SecLabelLabelIndexId, true,
							  SnapshotNow, 2, key);

	while (HeapTupleIsValid(tuple = systable_getnext(scan)))
		simple_heap_delete(rel, &tuple->t_self);

	systable_endscan(scan);

	heap_close(rel, RowExclusiveLock);
}

/*
 * XXX - it should be replaced by BKI scripts
 */
bool
seclabelCatalogHasSysAttr(Oid relOid)
{
	switch (relOid)
	{
		case AggregateRelationId:
			/* pg_aggregate is property of pg_proc */
		case AccessMethodOperatorRelationId:
			/* pg_amop is property of pg_opfamily */
		case AccessMethodProcedureRelationId:
			/* pg_amproc is property of pg_opfamily */
		case AttrDefaultRelationId:
			/* pg_attrdef is property of pg_attribute */
		case AuthMemRelationId:
			/* pg_auth_members is property of pg_auth */
		case ConstraintRelationId:
			/*
			 * CHECK constraint is property of pg_class
			 * DOMAIN constraint is property of pg_type
			 * Global assertion is property of pg_database
			 */
		case DbRoleSettingRelationId:
			/* pg_db_role_setting is property of pg_auth or pg_database */
		case DependRelationId:
			/* property of the depending object */
		case DescriptionRelationId:
			/* property of the object commented on */
		case EnumRelationId:
			/* pg_enum is property of pg_type */
		case IndexRelationId:
			/* pg_index is property of pg_class with RELKIND_INDEX */
		case InheritsRelationId:
			/* pg_inherits is property of the child relation */
		case LargeObjectRelationId:
			/* pg_largeobject is data chunk of pg_largeobject_metadata */
		case RewriteRelationId:
			/* pg_rewrite is property of pg_class */
		case SecLabelRelationId:
			/* No security attribute has no security label */
		case SharedDependRelationId:
			/* property of the depending shared object */
		case SharedDescriptionRelationId:
			/* property of the shared object commented on */
		case TriggerRelationId:
			/* pg_trigger is property of pg_class */
		case TSConfigMapRelationId:
			/* pg_ts_config_map is property of pg_ts_config */
			return false;

		default:
			return true;
	}
}

Oid *
seclabelMakeRelationDefaults(TupleDesc tupdesc, List *supOids)
{
	ListCell   *l;
	Oid		   *secLabels;
	Oid			securityId;
	int			index, attno, nitems;

	nitems = tupdesc->natts - FirstLowInvalidHeapAttributeNumber;
	secLabels = palloc0(sizeof(Oid) * nitems);

	foreach (l, supOids)
	{
		Oid		relOid = lfirst_oid(l);

		securityId = GetSysCacheSecid1(RELOID, ObjectIdGetDatum(relOid));

		if (!OidIsValid(secLabels[0]))
			secLabels[0] = securityId;
		else if (!seclabelCompareSecid(RelationRelationId, secLabels[0],
									   RelationRelationId, securityId))
			ereport(ERROR,
					(errcode(ERRCODE_WRONG_OBJECT_TYPE),
					 errmsg("cannot inherit relations with different label")));

		for (index = 1; index < nitems; index++)
		{
			attno = index + FirstLowInvalidHeapAttributeNumber;

			if (attno < 0)
				securityId = GetSysCacheSecid2(ATTNUM,
											   ObjectIdGetDatum(relOid),
											   Int16GetDatum(attno));
			else
			{
				const char *attname = NameStr(tupdesc->attrs[attno]->attname);
				securityId = GetSysCacheSecid2(ATTNAME,
											   ObjectIdGetDatum(relOid),
											   PointerGetDatum(attname));
			}

			if (!OidIsValid(securityId))
				continue;

			if (!OidIsValid(secLabels[index]))
				secLabels[index] = securityId;
			else if (!seclabelCompareSecid(AttributeRelationId, secLabels[index],
										   AttributeRelationId, securityId))
				ereport(ERROR,
						(errcode(ERRCODE_WRONG_OBJECT_TYPE),
						 errmsg("cannot inherit attribute with different label")));
		}
	}
	return secLabels;
}

Oid *
seclabelMakeToastDefaults(TupleDesc tupdesc, Oid relOid)
{
	Oid	   *secLabels;
	Oid		securityId;
	int		index, nitems;

	nitems = tupdesc->natts + 1 - FirstLowInvalidHeapAttributeNumber;
	secLabels = palloc0(sizeof(Oid) * nitems);

	securityId = GetSysCacheSecid1(RELOID, ObjectIdGetDatum(relOid));
	secLabels[0] = securityId;

	securityId = seclabelMoveSecid(AttributeRelationId,
								   RelationRelationId,
								   securityId);
	for (index = 1; index < nitems; index++)
		secLabels[index] = securityId;

	return secLabels;
}

Oid
seclabelGetNewSecid(Relation rel, HeapTuple tuple)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled() &&
		!IsBootstrapProcessingMode())
	{
		return sepgsql_get_default_secid(rel, tuple);
	}
#endif
	return InvalidOid;
}

static Oid
inputSecurityLabel(Oid relid, const char *seclabel)
{
	LOCKMODE		lockmode = AccessShareLock;
	Relation		rel;
	ScanKeyData		skey[3];
	SysScanDesc		scan;
	HeapTuple		tuple;
	Oid				datid;
	Oid				secid;
	Datum			values[Natts_pg_seclabel];
	bool			nulls[Natts_pg_seclabel];

	datid = (IsSharedRelation(relid) ? InvalidOid : MyDatabaseId);

retry:
	/*
	 * Lookup pg_seclabel first, then insert a new entry if not found.
	 * An exclusive lock is not necessary for the first read-only path,
	 * and we assume most of trials are read-only.
	 */
	rel = heap_open(SecLabelRelationId, lockmode);

	ScanKeyInit(&skey[0],
				Anum_pg_seclabel_datid,
				BTEqualStrategyNumber, F_OIDEQ,
				ObjectIdGetDatum(datid));
	ScanKeyInit(&skey[1],
				Anum_pg_seclabel_relid,
				BTEqualStrategyNumber, F_OIDEQ,
				ObjectIdGetDatum(relid));
	ScanKeyInit(&skey[2],
				Anum_pg_seclabel_label,
				BTEqualStrategyNumber, F_TEXTEQ,
				CStringGetTextDatum(seclabel));

	scan = systable_beginscan(rel, SecLabelLabelIndexId, true,
							  SnapshotToast, 3, skey);

	tuple = systable_getnext(scan);
	if (HeapTupleIsValid(tuple))
	{
		secid = ((Form_pg_seclabel) GETSTRUCT(tuple))->secid;
		systable_endscan(scan);
		heap_close(rel, lockmode);
		return secid;
	}

	/*
	 * If not exist, try to insert a new entry.
	 */
	if (lockmode == AccessShareLock)
	{
		systable_endscan(scan);
		heap_close(rel, lockmode);
		lockmode = RowExclusiveLock;
		goto retry;
	}

	memset(nulls, false, sizeof(nulls));
	secid = GetNewOidWithIndex(rel, SecLabelSecidIndexId,
							   Anum_pg_seclabel_secid);
	values[Anum_pg_seclabel_secid - 1] = ObjectIdGetDatum(secid);
	values[Anum_pg_seclabel_datid - 1] = ObjectIdGetDatum(datid);
	values[Anum_pg_seclabel_relid - 1] = ObjectIdGetDatum(relid);
	values[Anum_pg_seclabel_label - 1] = CStringGetTextDatum(seclabel);

	tuple = heap_form_tuple(RelationGetDescr(rel), values, nulls);

	simple_heap_insert(rel, tuple);

	CatalogUpdateIndexes(rel, tuple);

	systable_endscan(scan);

	heap_close(rel, lockmode);

	return secid;
}

static char *
outputSecurityLabel(Oid relid, Oid secid)
{
	Relation		rel;
	ScanKeyData		skey[3];
	SysScanDesc		scan;
	HeapTuple		tuple;
	Oid				datid;
	char		   *result = NULL;

	datid = (IsSharedRelation(relid) ? InvalidOid : MyDatabaseId);

	/*
	 * Lookup pg_seclabel for the given datid/relid/secid
	 */
	rel = heap_open(SecLabelRelationId, AccessShareLock);

	ScanKeyInit(&skey[0],
				Anum_pg_seclabel_secid,
				BTEqualStrategyNumber, F_OIDEQ,
				ObjectIdGetDatum(secid));
	ScanKeyInit(&skey[1],
				Anum_pg_seclabel_datid,
				BTEqualStrategyNumber, F_OIDEQ,
				ObjectIdGetDatum(datid));
	ScanKeyInit(&skey[2],
				Anum_pg_seclabel_relid,
				BTEqualStrategyNumber, F_OIDEQ,
				ObjectIdGetDatum(relid));

	scan = systable_beginscan(rel, SecLabelSecidIndexId, true,
							  SnapshotToast, 3, skey);
	tuple = systable_getnext(scan);
	if (HeapTupleIsValid(tuple))
	{
		Datum	datum;
		bool	isnull;

		datum = heap_getattr(tuple,
							 Anum_pg_seclabel_label,
							 RelationGetDescr(rel), &isnull);
		if (!isnull)
			result = TextDatumGetCString(datum);
	}
	systable_endscan(scan);

	heap_close(rel, AccessShareLock);

	return result;
}

Oid
seclabelRawInput(Oid relid, char *seclabel)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
		seclabel = sepgsql_rawlabel_in(seclabel);
#endif
	return inputSecurityLabel(relid, seclabel);
}

char *
seclabelRawOutput(Oid relid, Oid secid)
{
	char   *seclabel = outputSecurityLabel(relid, secid);
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
		seclabel = sepgsql_rawlabel_out(seclabel);
#endif
	return seclabel;
}

Oid
seclabelTransInput(Oid relid, char *seclabel)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
		seclabel = sepgsql_mcstrans_in(seclabel);
#endif
	return seclabelRawInput(relid, seclabel);
}

char *
seclabelTransOutput(Oid relid, Oid secid)
{
	char   *seclabel = seclabelRawOutput(relid, secid);
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
		seclabel = sepgsql_mcstrans_out(seclabel);
#endif
	return seclabel;
}

Oid
seclabelMoveSecid(Oid dst_relid, Oid src_relid, Oid secid)
{
	char   *label = seclabelRawOutput(src_relid, secid);

	if (!label)
		return InvalidOid;

	secid = seclabelRawInput(dst_relid, label);

	pfree(label);

	return secid;
}

bool
seclabelCompareSecid(Oid relid1, Oid secid1, Oid relid2, Oid secid2)
{
	char   *label1 = seclabelRawOutput(relid1, secid1);
	char   *label2 = seclabelRawOutput(relid2, secid2);
	bool	retval = false;

	if (label1 && label2 && strcmp(label1, label2) == 0)
		retval = true;
	else if (!label1 && !label2)
		retval = false;

	if (label1)
		pfree(label1);
	if (label2)
		pfree(label2);

	return retval;
}

Datum
seclabelSysattOutput(Oid relid, HeapTuple tuple)
{
	Oid		secid;
	char   *seclabel;

	secid = HeapTupleGetSecid(tuple);

	seclabel = seclabelTransOutput(relid, secid);
	if (!seclabel)
		seclabel = "unlabeled";

	return CStringGetTextDatum(seclabel);
}

/*
 * seclabelRelationReclaim
 *
 * It reclaims security labels already referenced to.
 * It has to be called under the VACUUM FULL context that means the relation
 * to be reclaimed is already locked exclusively.
 */
void
seclabelRelationReclaim(Oid relOid)
{
	StringInfoData	query;
	const char	   *nspname_reclaimed;
	const char	   *relname_reclaimed;
	const char	   *nspname_pg_seclabel;
	const char	   *relname_pg_seclabel;
	const char	   *attname_secid;
	const char	   *attname_datid;
	const char	   *attname_relid;
	const char	   *attname_label;
	const char	   *nspname_to_secid;
	const char	   *proname_to_secid;
	Oid				databaseId;
	Oid				namespaceId;
	int				index;
	int				save_sepgsql_mode;

	if (SPI_connect() != SPI_OK_CONNECT)
		elog(ERROR, "SPI_connect() failed");

	/*
	 * DELETE orphan entries
	 */
	databaseId = (IsSharedRelation(relOid) ? InvalidOid : MyDatabaseId);

	namespaceId = get_rel_namespace(relOid);
	nspname_reclaimed = get_namespace_name(namespaceId);
	relname_reclaimed = get_rel_name(relOid);

	namespaceId = get_rel_namespace(SecLabelRelationId);
	nspname_pg_seclabel = get_namespace_name(namespaceId);
	relname_pg_seclabel = get_rel_name(SecLabelRelationId);

	attname_secid = get_attname(SecLabelRelationId, Anum_pg_seclabel_secid);
	attname_datid = get_attname(SecLabelRelationId, Anum_pg_seclabel_datid);
	attname_relid = get_attname(SecLabelRelationId, Anum_pg_seclabel_relid);
	attname_label = get_attname(SecLabelRelationId, Anum_pg_seclabel_label);

	namespaceId = get_func_namespace(F_SECLABEL_TO_SECID);
	nspname_to_secid = get_namespace_name(namespaceId);
	proname_to_secid = get_func_name(F_SECLABEL_TO_SECID);

	initStringInfo(&query);
	appendStringInfo(&query,
					 "DELETE FROM %s.%s "
					 "WHERE %s = %u AND %s = %u AND %s NOT IN "
					 "(SELECT %s.%s(%s) FROM ONLY %s.%s) "
					 "RETURNING %s,%s",
					 quote_identifier(nspname_pg_seclabel),
					 quote_identifier(relname_pg_seclabel),
					 quote_identifier(attname_datid),
					 databaseId,
					 quote_identifier(attname_relid),
					 relOid,
					 quote_identifier(attname_secid),
					 quote_identifier(nspname_to_secid),
					 quote_identifier(proname_to_secid),
					 quote_identifier(relname_reclaimed),
					 quote_identifier(nspname_reclaimed),
					 quote_identifier(relname_reclaimed),
					 quote_identifier(attname_secid),
					 quote_identifier(attname_label));
	/*
	 * Run the query
	 */
	elog(DEBUG1, "query: %s", query.data);

	save_sepgsql_mode = sepostgresql_mode;

	PG_TRY();
	{
		sepostgresql_mode = SEPGSQL_MODE_INTERNAL;

		if (SPI_execute(query.data, false, 0) != SPI_OK_DELETE_RETURNING)
			elog(ERROR, "Failed to run: %s", query.data);
	}
	PG_CATCH();
    {
		sepostgresql_mode = save_sepgsql_mode;

		PG_RE_THROW();
	}
	PG_END_TRY();

	sepostgresql_mode = save_sepgsql_mode;

	for (index = 0; index < SPI_processed; index++)
	{
		HeapTuple	tuple = SPI_tuptable->vals[index];
		char	   *label;
		Oid			secid;
		Datum		datum;
		bool		isnull;

		datum = heap_getattr(tuple, 1, SPI_tuptable->tupdesc, &isnull);
		secid = (isnull ? InvalidOid : DatumGetObjectId(datum));

		datum = heap_getattr(tuple, 2, SPI_tuptable->tupdesc, &isnull);
		label = (isnull ? NULL : TextDatumGetCString(datum));

		elog(DEBUG1, "seclabel: \"%s\"was reclaimed (secid=%u)",
			 label, secid);
	}

	if (SPI_finish() != SPI_OK_FINISH)
		elog(ERROR, "SPI_finish() failed");
}

Datum
seclabel_to_secid(PG_FUNCTION_ARGS)
{
	HeapTupleHeader	htup = PG_GETARG_HEAPTUPLEHEADER(0);

	PG_RETURN_OID(HeapTupleHeaderGetSecid(htup));
}
