/*
 * attribute.c
 *
 * SELinux hooks related to attribute
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "access/heapam.h"
#include "access/sysattr.h"
#include "catalog/pg_attribute.h"
#include "catalog/pg_class.h"
#include "catalog/pg_seclabel.h"
#include "sepgsql/hooks.h"
#include "sepgsql/sepgsql.h"
#include "utils/lsyscache.h"
#include "utils/syscache.h"
#include "utils/tqual.h"

#ifdef HAVE_SELINUX
bool
sepgsql_attribute_common(Oid relOid, AttrNumber attno,
						 uint32 required, bool abort)
{
	Form_pg_attribute	attForm;
	HeapTuple		tuple;
	sepgsql_sid_t	tsid;
	char			auname[NAMEDATALEN * 2 + 10];
	bool			retval;

	Assert(get_rel_relkind(relOid) == RELKIND_RELATION);

	tuple = SearchSysCache2(ATTNUM,
							ObjectIdGetDatum(relOid),
							Int16GetDatum(attno));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for attribute %d of relation %u",
			 attno, relOid);
	attForm = (Form_pg_attribute) GETSTRUCT(tuple);

	tsid.relid = AttributeRelationId;
	tsid.secid = HeapTupleGetSecid(tuple);

	snprintf(auname, sizeof(auname), "%s.%s",
			 get_rel_name(relOid), NameStr(attForm->attname));

	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_COLUMN,
								  required,
								  auname,
								  abort);
	ReleaseSysCache(tuple);

	return retval;
}
#endif

Oid
sepgsql_attribute_create(Oid relOid, const char *attName)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;
		char	auname[NAMEDATALEN * 2 + 10];

		if (get_rel_relkind(relOid) != RELKIND_RELATION)
		{
			sepgsql_relation_common(relOid,
									SEPG_DB_TABLE__SETATTR, true);
			nsid.relid = RelationRelationId;
			nsid.secid = GetSysCacheSecid1(RELOID, ObjectIdGetDatum(relOid));

			nsid = sepgsql_move_secid(AttributeRelationId, nsid);

			return nsid.secid;
		}

		nsid = sepgsql_get_default_column_secid(relOid);
		snprintf(auname, sizeof(auname), "%s.%s",
				 get_rel_name(relOid), attName);

		/* db_column:{create} */
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_COLUMN,
							 SEPG_DB_COLUMN__CREATE,
							 auname,
							 true);
		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_attribute_alter(Oid relOid, const char *attName)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		AttrNumber		attnum = get_attnum(relOid, attName);

		if (attnum == InvalidAttrNumber)
			return;		/* to be failed later */

		if (get_rel_relkind(relOid) == RELKIND_RELATION)
			sepgsql_attribute_common(relOid, attnum,
									 SEPG_DB_COLUMN__SETATTR, true);
		else
			sepgsql_relation_common(relOid,
									SEPG_DB_TABLE__SETATTR, true);
	}
#endif
}

Oid
sepgsql_attribute_relabel(Oid relOid, const char *attName, char *new_label)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		AttrNumber		attnum = get_attnum(relOid, attName);
		sepgsql_sid_t	nsid;
		char	auname[NAMEDATALEN * 2 + 10];

		Assert(get_rel_relkind(relOid) == RELKIND_RELATION);
		if (attnum == InvalidAttrNumber)
			return InvalidOid;	/* to be failed later */

		nsid.relid = AttributeRelationId;
		nsid.secid = seclabelTransInput(nsid.relid, new_label);

		snprintf(auname, sizeof(auname), "%s.%s",
				 get_rel_name(relOid), get_attname(relOid, attnum));

		/* db_column:{setattr relabelfrom} */
		sepgsql_attribute_common(relOid, attnum,
								 SEPG_DB_COLUMN__SETATTR |
								 SEPG_DB_COLUMN__RELABELFROM,
								 true);

		/* db_column:{relabelto} */
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_COLUMN,
							 SEPG_DB_COLUMN__RELABELTO,
							 auname,
							 true);
		return nsid.secid;
	}
#endif
	ereport(ERROR,
			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
			 errmsg("SE-PostgreSQL is not available")));
	return InvalidOid;
}

void
sepgsql_attribute_drop(Oid relOid, const char *attName, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		AttrNumber	attnum = get_attnum(relOid, attName);

		/*
		 * If not found, the caller raises an error with an appropriate
		 * error message.
		 */
		if (attnum == InvalidAttrNumber)
			return;

		if (get_rel_relkind(relOid) == RELKIND_RELATION)
		{
			sepgsql_attribute_common(relOid, attnum,
									 SEPG_DB_COLUMN__DROP, true);
			/*
			 * ALTER TABLE SET WITHOUT SECURITY LABEL is equivalent to
			 * relabel all the tuples within the target relation.
			 * In this case, we need to check {relabelfrom relabelto}
			 */
			if (!cascade &&
				attnum == SecurityLabelAttributeNumber)
			{
				Relation		rel;
				HeapScanDesc	scan;
				HeapTuple		tuple;
				sepgsql_sid_t	tsid;

				rel = heap_open(relOid, AccessShareLock);

				scan = heap_beginscan(rel, SnapshotNow, 0, NULL);

				while (HeapTupleIsValid(tuple = heap_getnext(scan, ForwardScanDirection)))
				{
					/* db_tuple:{update relabelfrom} */
					tsid.relid = relOid;
					tsid.secid = HeapTupleGetSecid(tuple);

					sepgsql_client_perms(tsid,
										 SEPG_CLASS_DB_TUPLE,
										 SEPG_DB_TUPLE__UPDATE |
										 SEPG_DB_TUPLE__RELABELFROM,
										 NULL,
										 true);
				}
				heap_endscan(scan);

				heap_close(rel, AccessShareLock);

				/* db_tuple:{relabelto} */
				tsid.relid = RelationRelationId;
				tsid.secid = GetSysCacheSecid1(RELOID, ObjectIdGetDatum(relOid));

				sepgsql_client_perms(tsid,
									 SEPG_CLASS_DB_TUPLE,
									 SEPG_DB_TUPLE__RELABELTO,
									 NULL,
									 true);
			}
		}
		else if (!cascade)
			sepgsql_relation_common(relOid,
									SEPG_DB_TABLE__SETATTR, true);
	}
#endif
}

void
sepgsql_attribute_grant(Oid relOid, AttrNumber attnum)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		if (get_rel_relkind(relOid) == RELKIND_RELATION)
			sepgsql_attribute_common(relOid, attnum,
									 SEPG_DB_COLUMN__SETATTR, true);
		else
			sepgsql_relation_common(relOid,
									SEPG_DB_TABLE__SETATTR, true);
	}
#endif
}

void
sepgsql_attribute_comment(Oid relOid, AttrNumber attnum)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		if (get_rel_relkind(relOid) == RELKIND_RELATION)
			sepgsql_attribute_common(relOid, attnum,
									 SEPG_DB_COLUMN__SETATTR, true);
		else
			sepgsql_relation_common(relOid,
									SEPG_DB_TABLE__SETATTR, true);
	}
#endif
}
