/*
 * avc.c
 * userspace access vector cache
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "access/hash.h"
#include "access/xact.h"
#include "catalog/pg_seclabel.h"
#include "libpq/libpq-be.h"
#include "libpq/pqsignal.h"
#include "miscadmin.h"
#include "sepgsql/hooks.h"
#include "sepgsql/sepgsql.h"
#include "storage/shmem.h"
#include "storage/lock.h"
#include "utils/memutils.h"

#include <selinux/selinux.h>
#include <selinux/avc.h>
#include <unistd.h>

/* ------------------------------------------------------------
 *
 * Userspace Access Vector Cache
 *
 * ------------------------------------------------------------
 */
static MemoryContext AvcMemCtx = NULL;

#define AVC_HASH_NUM_SLOTS		256
#define AVC_HASH_NUM_NODES		180

#define avc_hash_key(trelid, tsecid, tclass, nrelid)	\
	(hash_uint32((trelid) ^ (tsecid) ^ ((tclass) << 3) ^ (nrelid)))

typedef struct _avc_datum
{
	uint32			hash_key;

	uint16			tclass;
	sepgsql_sid_t	tsid;
	sepgsql_sid_t	nsid;
	char		   *tcontext;
	char		   *ncontext;

	uint32			allowed;
	uint32			auditallow;
	uint32			auditdeny;
	bool			permissive;

	bool			hot_cache;
} avc_datum;

typedef struct _avc_page
{
	struct _avc_page   *next;

	List	   *slot[AVC_HASH_NUM_SLOTS];

	uint32		avc_count;
	uint32		lru_hint;

	char		scontext[1];
} avc_page;

static avc_page    *current_page = NULL;

static int			avc_version = -1;

/*
 * selinux_state
 *
 * It is deployed on the shared memory region, to show the system
 * state of SELinux and its security policy.
 */
struct
{
	int		version;

	bool	enforcing;
} *selinux_state = NULL;

Size
sepgsql_shmem_size(void)
{
	return sizeof(*selinux_state);
}

static void
sepgsql_shmem_init(void)
{
	bool	found;

	selinux_state = ShmemInitStruct("SELinux system state",
									sepgsql_shmem_size(), &found);
	if (!found)
	{
		LWLockAcquire(SepgsqlAvcLock, LW_EXCLUSIVE);

		selinux_state->version = 0;
		selinux_state->enforcing = (security_getenforce() > 0);

		LWLockRelease(SepgsqlAvcLock);
	}
}

void
sepgsql_avc_switch(const char *scontext)
{
	avc_page   *new_page;
	int			i, length;

	if (current_page)
	{
		new_page = current_page;
		do {
			if (strcmp(new_page->scontext, scontext) == 0)
			{
				current_page = new_page;
				return;
			}
			new_page = new_page->next;
		} while (new_page != current_page);
	}

	/*
	 * Not found, so create a new avc_page
	 */
	length = sizeof(avc_page) + strlen(scontext);
	new_page = MemoryContextAllocZero(AvcMemCtx, length);

	strcpy(new_page->scontext, scontext);
	for (i = 0; i < AVC_HASH_NUM_SLOTS; i++)
		new_page->slot[i] = NIL;

	if (!current_page)
		new_page->next = new_page;
	else
	{
		new_page->next = current_page->next;
		current_page->next = new_page;
	}
	current_page = new_page;
}


static void
sepgsql_avc_reset(void)
{
	Assert(AvcMemCtx != NULL);

	MemoryContextReset(AvcMemCtx);

	current_page = NULL;

	sepgsql_avc_switch(sepgsql_get_client_label());
}

static bool
sepgsql_avc_is_valid(void)
{
	bool	result = true;

	LWLockAcquire(SepgsqlAvcLock, LW_SHARED);
	if (avc_version != selinux_state->version)
	{
		sepgsql_avc_reset();

		/* copy current version to local variable */
		avc_version = selinux_state->version;

		result = false;
	}
	LWLockRelease(SepgsqlAvcLock);

	return result;
}


static void
sepgsql_avc_reclaim(avc_page *page)
{
	ListCell   *l;

	while (page->avc_count >  AVC_HASH_NUM_NODES - 10)
	{
		foreach (l, page->slot[page->lru_hint])
		{
			avc_datum  *cache = lfirst(l);

			if (cache->hot_cache)
				cache->hot_cache = false;
			{
				list_delete_ptr(page->slot[page->lru_hint], cache);
				pfree(cache);
				page->avc_count--;
			}
		}
		page->lru_hint = (page->lru_hint + 1) % AVC_HASH_NUM_SLOTS;
	}
}

static avc_datum *
sepgsql_avc_make_entry(avc_page *page,
					   sepgsql_sid_t tsid, uint16 tclass, Oid nrelid)
{
	struct av_decision	avd;
	MemoryContext	oldctx;
	char		   *scontext;
	char		   *tcontext;
	char		   *ncontext;
	avc_datum	   *cache;
	uint32			hash_key, index;

	hash_key = avc_hash_key(tsid.relid, tsid.secid, tclass, nrelid);
	index = hash_key % AVC_HASH_NUM_SLOTS;

	oldctx = MemoryContextSwitchTo(AvcMemCtx);

    scontext = page->scontext;
	tcontext = seclabelRawOutput(tsid.relid, tsid.secid);
	ncontext = sepgsql_compute_create(scontext, tcontext, tclass);

	sepgsql_compute_avd(scontext, tcontext, tclass, &avd);

    cache = palloc0(sizeof(avc_datum));

    cache->hash_key = hash_key;

    cache->tclass = tclass;

    cache->hot_cache = true;
    cache->tcontext = tcontext;
    cache->ncontext = ncontext;

    cache->tsid.relid = tsid.relid;
    cache->tsid.secid = tsid.secid;

	if (OidIsValid(nrelid))
	{
		cache->nsid.relid = nrelid;
		cache->nsid.secid = seclabelRawInput(nrelid, ncontext);
	}

	cache->allowed = avd.allowed;
	cache->auditallow = avd.auditallow;
	cache->auditdeny = avd.auditdeny;
	if (avd.flags & SELINUX_AVD_FLAGS_PERMISSIVE)
		cache->permissive = true;

	if (page->avc_count > AVC_HASH_NUM_NODES)
		sepgsql_avc_reclaim(page);

	page->slot[index] = lcons(cache, page->slot[index]);
	page->avc_count++;

	MemoryContextSwitchTo(oldctx);

	return cache;
}

static avc_datum *
sepgsql_avc_lookup(avc_page *page,
				   sepgsql_sid_t tsid, uint16 tclass, Oid nrelid)
{
    avc_datum  *cache = NULL;
    uint32      hash_key, index;
    ListCell   *l;

    hash_key = avc_hash_key(tsid.relid, tsid.secid, tclass, nrelid);
    index = hash_key % AVC_HASH_NUM_SLOTS;

    foreach (l, page->slot[index])
    {
        cache = lfirst(l);
        if (cache->hash_key == hash_key &&
            cache->tclass == tclass &&
            cache->tsid.relid == tsid.relid &&
            cache->tsid.secid == tsid.secid &&
            cache->nsid.relid == nrelid)
        {
            cache->hot_cache = true;
            return cache;
        }
    }
    return NULL;
}

bool
sepgsql_client_perms(sepgsql_sid_t tsid,
					 uint16 tclass, uint32 required,
					 const char *audit_name, bool abort)
{
	avc_datum  *cache;
	uint32		denied, audited;
	bool		result = true;

	do {
		cache = sepgsql_avc_lookup(current_page,
								   tsid, tclass, InvalidOid);
		if (!cache)
			cache = sepgsql_avc_make_entry(current_page,
										   tsid, tclass, InvalidOid);
	} while (!sepgsql_avc_is_valid());

	denied = required & ~cache->allowed;
	if (sepgsql_debug_audit && tclass != SEPG_CLASS_DB_TUPLE)
		audited = (denied ? (denied & ~0) : (required & ~0));
	else
		audited = (denied ? (denied & cache->auditdeny)
						  : (required & cache->auditallow));

	if (audited)
	{
		sepgsql_audit_log(!!denied,
						  current_page->scontext,
						  seclabelRawOutput(tsid.relid, tsid.secid),
						  tclass, audited, audit_name);
	}

	if (denied)
	{
		if (!sepgsql_get_enforce() || cache->permissive)
			cache->allowed |= required;		/* prevent flood of audit log */
		else
		{
			if (abort)
				ereport(ERROR,
						(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
						 errmsg("SELinux: security policy violation")));
			result = false;
		}
	}

    return result;
}

sepgsql_sid_t
sepgsql_client_create_secid(sepgsql_sid_t tsid, uint16 tclass, Oid nrelid)
{
	avc_datum	   *cache;

	do {
		cache = sepgsql_avc_lookup(current_page, tsid, tclass, nrelid);

		if (!cache)
			cache = sepgsql_avc_make_entry(current_page,
										   tsid, tclass, nrelid);
	} while (!sepgsql_avc_is_valid());

	return cache->nsid;
}

char *
sepgsql_client_create_label(sepgsql_sid_t tsid, uint16 tclass)
{
	avc_datum	   *cache;

	do {
		cache = sepgsql_avc_lookup(current_page, tsid, tclass, InvalidOid);

		if (!cache)
			cache = sepgsql_avc_make_entry(current_page,
										   tsid, tclass, InvalidOid);
	} while (!sepgsql_avc_is_valid());

	return cache->ncontext;
}

static void
sepgsql_avc_xact_callback(XactEvent event, void *arg)
{
	if (event == XACT_EVENT_ABORT)
		sepgsql_avc_reset();
}

static void
sepgsql_avc_sub_xact_callback(SubXactEvent event, SubTransactionId mySubid,
							  SubTransactionId parentSubid, void *arg)
{
	if (event == SUBXACT_EVENT_ABORT_SUB)
		sepgsql_avc_reset();
}

void
sepgsql_avc_init(void)
{
	sepgsql_shmem_init();

	AvcMemCtx = AllocSetContextCreate(TopMemoryContext,
									  "Userspace AVC",
									  ALLOCSET_DEFAULT_MINSIZE,
									  ALLOCSET_DEFAULT_INITSIZE,
									  ALLOCSET_DEFAULT_MAXSIZE);
	/*
	 * userspace avc should be invalidate when the current transaction
	 * is aborted on errors, because sid to be created shall be rollbacked.
	 */
	RegisterXactCallback(sepgsql_avc_xact_callback, NULL);
	RegisterSubXactCallback(sepgsql_avc_sub_xact_callback, NULL);
}

/* ------------------------------------------------------------
 *
 * SELinux state monitor process
 *
 * ------------------------------------------------------------
 */
static int
sepgsql_cb_log(int type, const char *fmt, ...)
{
	char	   *c, buffer[1024];
	va_list		ap;

	va_start(ap, fmt);
	vsnprintf(buffer, sizeof(buffer), fmt, ap);
	va_end(ap);

	c = strrchr(buffer, '\n');
	if (c)
		*c = '\0';

	ereport(LOG,(errmsg("%s", buffer)));

	return 0;
}

static int
sepgsql_cb_setenforce(int enforce)
{
	LWLockAcquire(SepgsqlAvcLock, LW_EXCLUSIVE);
	selinux_state->enforcing = (enforce ? true : false);
	selinux_state->version++;
	LWLockRelease(SepgsqlAvcLock);

	return 0;
}

static int
sepgsql_cb_policyload(int seqno)
{
	LWLockAcquire(SepgsqlAvcLock, LW_EXCLUSIVE);
	selinux_state->version++;
	LWLockRelease(SepgsqlAvcLock);

	return 0;
}

void
sepgsql_avc_worker_main(void)
{
	union selinux_callback cb;

	Assert(sepgsql_is_enabled());

#ifdef HAVE_SETSID
	if (setsid() < 0)
		elog(FATAL, "setsid() failed: %m");
#endif

	/*
	 * setup the signal handler
	 */
	pqinitmask();
	pqsignal(SIGHUP, SIG_IGN);
	pqsignal(SIGINT, SIG_IGN);
	pqsignal(SIGTERM, exit);
	pqsignal(SIGQUIT, exit);
	pqsignal(SIGUSR1, SIG_IGN);
	pqsignal(SIGUSR2, SIG_IGN);
	pqsignal(SIGCHLD, SIG_DFL);
	PG_SETMASK(&UnBlockSig);

	/*
	 * map shared memory segment
	 */
	sepgsql_shmem_init();

	ereport(LOG, (errmsg("SELinux: netlink receiver (pid=%u)", getpid())));

	/*
	 * setup callback functions from avc_netlink_loop()
	 */
	cb.func_log = sepgsql_cb_log;
	selinux_set_callback(SELINUX_CB_LOG, cb);
	cb.func_setenforce = sepgsql_cb_setenforce;
	selinux_set_callback(SELINUX_CB_SETENFORCE, cb);
	cb.func_policyload = sepgsql_cb_policyload;
	selinux_set_callback(SELINUX_CB_POLICYLOAD, cb);

	/*
	 * open netlink socket and wait for messages
	 */
	avc_netlink_open(1);

	avc_netlink_loop();

	exit(0);
}
