/*
 * blob.c
 *
 * SELinux hooks related to large objects
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "access/genam.h"
#include "access/heapam.h"
#include "access/sysattr.h"
#include "catalog/indexing.h"
#include "catalog/pg_largeobject.h"
#include "catalog/pg_largeobject_metadata.h"
#include "catalog/pg_seclabel.h"
#include "miscadmin.h"
#include "sepgsql/hooks.h"
#include "sepgsql/sepgsql.h"
#include "utils/fmgroids.h"
#include "utils/tqual.h"

#ifdef HAVE_SELINUX
bool
sepgsql_largeobject_common(Oid loid, Snapshot snapshot,
						   uint32 required, bool abort)
{
	Relation		pg_lo_meta;
	ScanKeyData		skey;
	SysScanDesc		scan;
	HeapTuple		tuple;
	sepgsql_sid_t	tsid;
	char			auname[NAMEDATALEN];
	bool			retval;

	snprintf(auname, sizeof(auname), "blob_%u", loid);

	pg_lo_meta = heap_open(LargeObjectMetadataRelationId,
						   AccessShareLock);

	ScanKeyInit(&skey,
				ObjectIdAttributeNumber,
				BTEqualStrategyNumber, F_OIDEQ,
				ObjectIdGetDatum(loid));

	scan = systable_beginscan(pg_lo_meta,
							  LargeObjectMetadataOidIndexId, true,
							  snapshot, 1, &skey);

	tuple = systable_getnext(scan);
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "large object %u does not exist", loid);

	tsid.relid = LargeObjectMetadataRelationId;
	tsid.secid = HeapTupleGetSecid(tuple);

	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_BLOB,
								  required,
								  auname,
								  abort);
	systable_endscan(scan);

    heap_close(pg_lo_meta, AccessShareLock);

	return retval;
}
#endif

Oid
sepgsql_largeobject_create(Oid loid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;
		char			auname[NAMEDATALEN];

		nsid = sepgsql_get_default_blob_secid(MyDatabaseId);
		snprintf(auname, sizeof(auname), "blob_%u", loid);

		/* db_blob:{create} */
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_BLOB,
							 SEPG_DB_BLOB__CREATE,
							 auname,
							 true);
		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_largeobject_alter(Oid loid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_largeobject_common(loid, SnapshotNow,
								   SEPG_DB_BLOB__SETATTR, true);
	}
#endif
}

Oid
sepgsql_largeobject_relabel(Oid loid, char *newLabel)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;
		char			auname[NAMEDATALEN];

		nsid.relid = LargeObjectMetadataRelationId;
		nsid.secid = seclabelTransInput(nsid.relid, newLabel);

		snprintf(auname, sizeof(auname), "blob_%u", loid);

		/* db_blob:{setattr relabelfrom} */
		sepgsql_largeobject_common(loid,
								   SnapshotNow,
								   SEPG_DB_BLOB__SETATTR |
								   SEPG_DB_BLOB__RELABELFROM,
								   true);
		/* db_blob:{relabelto} */
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_BLOB,
							 SEPG_DB_BLOB__RELABELTO,
							 auname,
							 true);
		return nsid.secid;
	}
#endif
	ereport(ERROR,
			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
			 errmsg("SE-PostgreSQL is not available")));
	return InvalidOid;
}

void
sepgsql_largeobject_drop(Oid loid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_largeobject_common(loid, SnapshotNow,
								   SEPG_DB_BLOB__SETATTR, true);
	}
#endif
}

void
sepgsql_largeobject_read(Oid loid, Snapshot snapshot)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_largeobject_common(loid, snapshot,
								   SEPG_DB_BLOB__READ, true);
	}
#endif
}

void
sepgsql_largeobject_write(Oid loid, Snapshot snapshot)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_largeobject_common(loid, snapshot,
								   SEPG_DB_BLOB__WRITE, true);
	}
#endif
}

Oid
sepgsql_largeobject_import(Oid loid, const char *filename)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;
		char			auname[NAMEDATALEN];

		nsid = sepgsql_get_default_blob_secid(MyDatabaseId);
		snprintf(auname, sizeof(auname), "blob_%u", loid);

		/* db_blob:{create} */
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_BLOB,
							 SEPG_DB_BLOB__CREATE |
							 SEPG_DB_BLOB__WRITE |
							 SEPG_DB_BLOB__IMPORT,
							 auname,
							 true);
		/* db_file:{read} */
		// XXX - todo: add file read checks

		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_largeobject_export(Oid loid, Snapshot snapshot, const char *filename)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_blob:{read export} */
		sepgsql_largeobject_common(loid, snapshot,
								   SEPG_DB_BLOB__READ |
								   SEPG_DB_BLOB__EXPORT, true);
		/* file:{write} */
		// TODO: add security checks
	}
#endif
}

void
sepgsql_largeobject_grant(Oid loid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_largeobject_common(loid, SnapshotNow,
								   SEPG_DB_BLOB__SETATTR, true);
	}
#endif
}

void
sepgsql_largeobject_comment(Oid loid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_largeobject_common(loid, SnapshotNow,
								   SEPG_DB_BLOB__SETATTR, true);
	}
#endif
}
