/*
 * conversion.c
 *
 * SELinux hooks related to conversion
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "catalog/pg_conversion.h"
#include "sepgsql/hooks.h"
#include "sepgsql/sepgsql.h"
#include "utils/syscache.h"

#ifdef HAVE_SELINUX
bool
sepgsql_conversion_common(Oid convOid, uint32 required, bool abort)
{
	sepgsql_sid_t	tsid;
	HeapTuple		tuple;
	const char	   *auname;
	bool			retval;

	tuple = SearchSysCache1(CONVOID, ObjectIdGetDatum(convOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for conversion %u", convOid);

	tsid.relid = ConversionRelationId;
	tsid.secid = HeapTupleGetSecid(tuple);

	auname = NameStr(((Form_pg_conversion) GETSTRUCT(tuple))->conname);

	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_TUPLE,
								  required,
								  auname,
								  abort);
	ReleaseSysCache(tuple);

	return retval;
}

static Oid
get_conversion_namespace(Oid convOid)
{
	HeapTuple		tuple;
	Oid				namespaceId = InvalidOid;

	tuple = SearchSysCache1(CONVOID, ObjectIdGetDatum(convOid));
	if (HeapTupleIsValid(tuple))
	{
		namespaceId = ((Form_pg_conversion) GETSTRUCT(tuple))->connamespace;

		ReleaseSysCache(tuple);
	}
	return namespaceId;
}
#endif

Oid
sepgsql_conversion_create(const char *convName,
						  Oid namespaceId, Oid conversionFunc)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;

		nsid = sepgsql_get_default_tuple_secid(ConversionRelationId);

		/* db_schema:{add_name} */
		sepgsql_schema_common(namespaceId,
							  SEPG_DB_SCHEMA__ADD_NAME, true);
		/* db_procedure:{install} */
		sepgsql_proc_common(conversionFunc,
							SEPG_DB_PROCEDURE__INSTALL, true);
		/* db_tuple:{insert} */
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_TUPLE,
							 SEPG_DB_TUPLE__INSERT,
							 convName,
							 true);
		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_conversion_alter(Oid convOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{update} */
		sepgsql_conversion_common(convOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_conversion_alter_rename(Oid convOid, const char *newName)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		namespaceId = get_conversion_namespace(convOid);

		/* db_schema:{add_name remove_name} */
		sepgsql_schema_common(namespaceId,
							  SEPG_DB_SCHEMA__ADD_NAME |
							  SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_tuple:{update} */
		sepgsql_conversion_common(convOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_conversion_drop(Oid convOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		namespaceId = get_conversion_namespace(convOid);

		/* db_schema:{remove_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_tuple:{delete} */
		sepgsql_conversion_common(convOid, SEPG_DB_TUPLE__DELETE, true);
	}
#endif
}

void
sepgsql_conversion_comment(Oid convOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_conversion_common(convOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}
