/*
 * database.c
 *
 * SELinux hooks related to database
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "catalog/pg_database.h"
#include "catalog/pg_seclabel.h"
#include "commands/dbcommands.h"
#include "miscadmin.h"
#include "sepgsql/hooks.h"
#include "sepgsql/sepgsql.h"
#include "utils/syscache.h"

#ifdef HAVE_SELINUX
bool
sepgsql_database_common(Oid datOid, uint32 required, bool abort)
{
	HeapTuple		tuple;
	sepgsql_sid_t	tsid;
	const char	   *auname;
	bool			retval;

	tuple = SearchSysCache1(DATABASEOID,
							ObjectIdGetDatum(datOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for database %u", datOid);

	tsid.relid = DatabaseRelationId;
	tsid.secid = HeapTupleGetSecid(tuple);

	auname = NameStr(((Form_pg_database) GETSTRUCT(tuple))->datname);
	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_DATABASE,
								  required,
								  auname,
								  abort);
	ReleaseSysCache(tuple);

	return retval;
}
#endif

Oid
sepgsql_database_create(const char *datName, Oid templateOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;

		/* compute a default security context */
		nsid = sepgsql_get_default_database_secid(templateOid);

		/* db_database:{create} */
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_DATABASE,
							 SEPG_DB_DATABASE__CREATE,
							 datName, true);
		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_database_alter(Oid databaseOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_database_common(databaseOid,
								SEPG_DB_DATABASE__SETATTR,
								true);
	}
#endif
}

Oid
sepgsql_database_relabel(Oid databaseOid, char *new_label)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;
		char		   *auname;

		nsid.relid = DatabaseRelationId;
		nsid.secid = seclabelTransInput(nsid.relid, new_label);

		auname = get_database_name(databaseOid);

		/* db_database:{setattr relabelfrom} */
		sepgsql_database_common(databaseOid,
								SEPG_DB_DATABASE__SETATTR |
								SEPG_DB_DATABASE__RELABELFROM,
								true);

		/* db_database:{relabelto} */
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_DATABASE,
							 SEPG_DB_DATABASE__RELABELTO,
							 auname,
							 true);
		pfree(auname);

		return nsid.secid;
	}
#endif
	ereport(ERROR,
			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
			 errmsg("SE-PostgreSQL is not available")));
	return InvalidOid;
}

void
sepgsql_database_drop(Oid databaseOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_database_common(databaseOid,
								SEPG_DB_DATABASE__DROP,
								true);
	}
#endif
}

void
sepgsql_database_grant(Oid databaseOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_database_common(databaseOid,
                                SEPG_DB_DATABASE__SETATTR,
								true);
	}
#endif
}

void
sepgsql_database_comment(Oid databaseOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_database_common(databaseOid,
								SEPG_DB_DATABASE__SETATTR,
								true);
	}
#endif
}

void
sepgsql_database_connect(Oid databaseOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		if (!sepgsql_database_common(databaseOid,
									 SEPG_DB_DATABASE__ACCESS,
									 false))
			ereport(FATAL,
					(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
					 errmsg("permission denied for database \"%s\"",
							get_database_name(databaseOid))));
	}
#endif
}

void
sepgsql_database_reindex(Oid databaseOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		// TODO: check db_table:{indexon} for each



	}
#endif
}

void
sepgsql_database_getattr(Oid databaseOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_database_common(databaseOid,
								SEPG_DB_DATABASE__GETATTR,
								true);
	}
#endif
}
