/*
 * fdw.c
 *
 * SELinux hooks related to foreign data wrapper
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "catalog/pg_foreign_data_wrapper.h"
#include "catalog/pg_foreign_server.h"
#include "catalog/pg_user_mapping.h"
#include "miscadmin.h"
#include "sepgsql/hooks.h"
#include "sepgsql/sepgsql.h"
#include "utils/syscache.h"

#ifdef HAVE_SELINUX
bool
sepgsql_fdw_common(Oid fdwOid, uint32 required, bool abort)
{
	Form_pg_foreign_data_wrapper	fdwForm;
	HeapTuple		tuple;
	sepgsql_sid_t	tsid;
	bool			retval;

	tuple = SearchSysCache1(FOREIGNDATAWRAPPEROID, ObjectIdGetDatum(fdwOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for foreign-data-wrapper %u", fdwOid);
	fdwForm = (Form_pg_foreign_data_wrapper) GETSTRUCT(tuple);

	tsid.relid = ForeignDataWrapperRelationId;
	tsid.secid = HeapTupleGetSecid(tuple);

	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_TUPLE,
								  required,
								  NameStr(fdwForm->fdwname),
								  abort);
	ReleaseSysCache(tuple);

	return retval;
}

bool
sepgsql_fserver_common(Oid fservOid, uint32 required, bool abort)
{
	HeapTuple		tuple;
	sepgsql_sid_t	tsid;
	const char	   *auname;
	bool			retval;

	tuple = SearchSysCache1(FOREIGNSERVEROID, ObjectIdGetDatum(fservOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for foreign-server %u", fservOid);

	tsid.relid = ForeignServerRelationId;
	tsid.secid = HeapTupleGetSecid(tuple);

	auname = NameStr(((Form_pg_foreign_server) GETSTRUCT(tuple))->srvname);
	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_TUPLE,
								  required,
								  auname,
								  abort);
	ReleaseSysCache(tuple);

	return retval;
}

static char *
get_fserver_name(Oid fservOid)
{
	Form_pg_foreign_server	servForm;
	HeapTuple	tuple;
	char	   *srvname = NULL;

	tuple = SearchSysCache1(FOREIGNSERVEROID, ObjectIdGetDatum(fservOid));
	if (HeapTupleIsValid(tuple))
	{
		servForm = (Form_pg_foreign_server) GETSTRUCT(tuple);

		srvname = pstrdup(NameStr(servForm->srvname));

		ReleaseSysCache(tuple);
	}
	return srvname;
}

bool
sepgsql_user_mapping_common(Oid umapOid, uint32 required, bool abort)
{
	HeapTuple		tuple;
	sepgsql_sid_t	tsid;
	Oid				umuser;
	Oid				umserver;
	char			auname[NAMEDATALEN * 2 + 10];
	bool			retval;

	tuple = SearchSysCache1(USERMAPPINGOID, ObjectIdGetDatum(umapOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for user mapping %u", umapOid);

	tsid.relid = UserMappingRelationId;
	tsid.secid = HeapTupleGetSecid(tuple);

	umuser = ((Form_pg_user_mapping) GETSTRUCT(tuple))->umuser;
	umserver = ((Form_pg_user_mapping) GETSTRUCT(tuple))->umserver;

	snprintf(auname, sizeof(auname), "%s@%s",
			 OidIsValid(umuser) ? GetUserNameFromId(umuser) : "public",
			 get_fserver_name(umserver));

	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_TUPLE,
								  required,
								  auname,
								  abort);
	ReleaseSysCache(tuple);

	return retval;
}
#endif

Oid
sepgsql_fdw_create(const char *fdwName, Oid validatorFunc)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid
			= sepgsql_get_default_tuple_secid(ForeignDataWrapperRelationId);

		/* db_procedure:{install} */
		if (OidIsValid(validatorFunc))
			sepgsql_proc_common(validatorFunc,
								SEPG_DB_PROCEDURE__INSTALL, true);

		/* db_tuple:{insert} */
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_TUPLE,
							 SEPG_DB_TUPLE__INSERT,
							 fdwName,
							 true);
		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_fdw_alter(Oid fdwOid, Oid newValidator)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_procedure:{install} */
		if (OidIsValid(newValidator))
			sepgsql_proc_common(newValidator,
								SEPG_DB_PROCEDURE__INSTALL, true);
		/* db_tuple:{update} */
		sepgsql_fdw_common(fdwOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_fdw_drop(Oid fdwOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{delete} */
		sepgsql_fdw_common(fdwOid, SEPG_DB_TUPLE__DELETE, true);
	}
#endif
}

void
sepgsql_fdw_grant(Oid fdwOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{update} */
		sepgsql_fdw_common(fdwOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

Oid
sepgsql_fserver_create(const char *fservName, Oid fdwOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid
			= sepgsql_get_default_tuple_secid(ForeignServerRelationId);

		/* db_tuple:{insert} */
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_TUPLE,
							 SEPG_DB_TUPLE__INSERT,
							 fservName,
							 true);
		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_fserver_alter(Oid fservOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{update} */
		sepgsql_fserver_common(fservOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_fserver_drop(Oid fservOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{update} */
		sepgsql_fserver_common(fservOid, SEPG_DB_TUPLE__DELETE, true);
	}
#endif
}

void
sepgsql_fserver_grant(Oid fservOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{update} */
		sepgsql_fserver_common(fservOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

Oid
sepgsql_user_mapping_create(Oid umuserId, Oid fservOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		char			auname[NAMEDATALEN * 2 + 10];
		sepgsql_sid_t	nsid
			= sepgsql_get_default_tuple_secid(UserMappingRelationId);

		snprintf(auname, sizeof(auname), "%s@%s",
				 OidIsValid(umuserId) ? GetUserNameFromId(umuserId) : "public",
				 get_fserver_name(fservOid));

		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_TUPLE,
							 SEPG_DB_TUPLE__INSERT,
							 auname,
							 true);
		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_user_mapping_alter(Oid umapOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{update} */
		sepgsql_user_mapping_common(umapOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_user_mapping_drop(Oid umapOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{delete} */
		sepgsql_user_mapping_common(umapOid, SEPG_DB_TUPLE__DELETE, true);
	}
#endif
}
