/*
 * file.c
 *
 * SELinux hooks related to filesystem
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "sepgsql/hooks.h"
#include "sepgsql/sepgsql.h"

#include <libgen.h>
#ifdef HAVE_SELINUX
#include <selinux/selinux.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#ifdef HAVE_SELINUX
static uint16
sepgsql_file_class(const char *filename, bool abort)
{
	struct stat	stbuf;
	uint16		tclass;

	if (stat(filename, &stbuf) < 0)
	{
		if (abort)
			ereport(ERROR,
					(errcode_for_file_access(),
					 errmsg("could not stat file \"%s\": %m", filename)));
		return 0;
	}

	/* decide object class to be applied */
	if (S_ISDIR(stbuf.st_mode))
		tclass = SEPG_CLASS_DIR;
	else if (S_ISCHR(stbuf.st_mode))
		tclass = SEPG_CLASS_CHR_FILE;
	else if (S_ISBLK(stbuf.st_mode))
		tclass = SEPG_CLASS_BLK_FILE;
	else if (S_ISFIFO(stbuf.st_mode))
		tclass = SEPG_CLASS_FIFO_FILE;
	else if (S_ISLNK(stbuf.st_mode))
		tclass = SEPG_CLASS_LNK_FILE;
	else if (S_ISSOCK(stbuf.st_mode))
		tclass = SEPG_CLASS_SOCK_FILE;
	else
		tclass = SEPG_CLASS_FILE;

	return tclass;
}

static bool
sepgsql_file_common(const char *filename, uint16 tclass, uint32 required, bool abort)
{
	struct stat	stbuf;
	char	   *scontext = sepgsql_get_client_label();
	char	   *tcontext;
	char	   *temp;
	bool		retval;

	if (stat(filename, &stbuf) < 0)
		ereport(ERROR,
				(errcode_for_file_access(),
				 errmsg("could not stat file \"%s\"", filename)));

	if (getfilecon_raw(filename, &temp) < 0)
		ereport(ERROR,
				(errcode_for_file_access(),
				 errmsg("could not get security context \"%s\"", filename)));

	PG_TRY();
	{
		tcontext = pstrdup(temp);
	}
	PG_CATCH();
	{
		freecon(temp);
		PG_RE_THROW();
	}
	PG_END_TRY();
	freecon(temp);

	retval = sepgsql_compute_perms(scontext, tcontext,
								   tclass, required,
								   filename, abort);
	pfree(tcontext);

	return retval;
}
#endif

void
sepgsql_file_getattr(const char *filename)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		uint16	tclass = sepgsql_file_class(filename, true);

		sepgsql_file_common(filename, tclass,
							SEPG_FILE__GETATTR, true);
	}
#endif
}

void
sepgsql_file_read(const char *filename)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		uint16	tclass = sepgsql_file_class(filename, true);

		sepgsql_file_common(filename, tclass,
							SEPG_FILE__READ, true);
	}
#endif
}

void
sepgsql_file_write(const char *filename, bool may_create)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		uint16	tclass = sepgsql_file_class(filename, !may_create);
		char   *dir_name;
		char   *tcontext;
		char   *ncontext;
		char   *temp;

		/* write an existing file */
		if (tclass != 0)
		{
			sepgsql_file_common(filename, tclass,
								SEPG_FILE__WRITE, true);
			return;
		}

		/* create a new file, and write it */
		dir_name = dirname(pstrdup(filename));

		if (sepgsql_file_class(dir_name, true) != SEPG_CLASS_DIR)
			ereport(ERROR,
					(errcode(ERRCODE_WRONG_OBJECT_TYPE),
					 errmsg("\"%s\" is not a directory", dir_name)));

		if (getfilecon(dir_name, &temp) < 0)
			ereport(ERROR,
					(errcode_for_file_access(),
					 errmsg("could not get security context \"%s\"", dir_name)));
		PG_TRY();
		{
			tcontext = pstrdup(temp);
		}
		PG_CATCH();
		{
			freecon(temp);
			PG_RE_THROW();
		}
		PG_END_TRY();
		freecon(temp);

		/* dir:{add_name} */
		sepgsql_file_common(dir_name, SEPG_CLASS_DIR,
							SEPG_DIR__ADD_NAME, true);

		/* file:{create write} */
		ncontext = sepgsql_compute_create(sepgsql_get_client_label(),
										  tcontext, SEPG_CLASS_FILE);
		sepgsql_compute_perms(sepgsql_get_client_label(),
							  ncontext,
							  SEPG_CLASS_FILE,
							  SEPG_FILE__CREATE |
							  SEPG_FILE__WRITE,
							  filename, true);
	}
#endif
}

void
sepgsql_file_append(const char *filename)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		if (sepgsql_file_class(filename, true) != SEPG_CLASS_FILE)
			ereport(ERROR,
					(errcode(ERRCODE_WRONG_OBJECT_TYPE),
					 errmsg("\"%s\" is not a regular file", filename)));

		sepgsql_file_common(filename, SEPG_CLASS_FILE,
							SEPG_FILE__APPEND, true);
	}
#endif
}

void
sepgsql_file_unlink(const char *filename)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		uint16	tclass = sepgsql_file_class(filename, true);

		sepgsql_file_common(filename, tclass, SEPG_FILE__UNLINK, true);
	}
#endif
}

void
sepgsql_file_rename(const char *oldpath, const char *newpath)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		char	   *olddir = dirname(pstrdup(oldpath));
		char	   *newdir = dirname(pstrdup(newpath));
		uint16		tclass;
		uint32		required;

		/* file:{setattr} on oldpath */
		tclass = sepgsql_file_class(oldpath, true);
		sepgsql_file_common(oldpath, tclass,
							SEPG_FILE__RENAME, true);

		if (sepgsql_file_class(olddir, true) != SEPG_CLASS_DIR)
			ereport(ERROR,
					(errcode(ERRCODE_WRONG_OBJECT_TYPE),
					 errmsg("\"%s\" is not a directory", olddir)));

		required = SEPG_DIR__SEARCH | SEPG_DIR__REMOVE_NAME;
		if (strcmp(olddir, newdir) == 0)
		{
			/* dir:{search remove_name add_name} */
			required |= SEPG_DIR__ADD_NAME;
			sepgsql_file_common(olddir, SEPG_CLASS_DIR, required, true);
		}
		else
		{
			/* dir:{search remove_name (reparent)} */
			if (tclass == SEPG_CLASS_DIR)
				required |= SEPG_DIR__REPARENT;
			sepgsql_file_common(olddir, SEPG_CLASS_DIR, required, true);

			if (sepgsql_file_class(newdir, true) != SEPG_CLASS_DIR)
				ereport(ERROR,
						(errcode(ERRCODE_WRONG_OBJECT_TYPE),
						 errmsg("\"%s\" is not a directory", newdir)));

			/* dir:{add_name} */
			sepgsql_file_common(newdir, SEPG_CLASS_DIR,
								SEPG_DIR__ADD_NAME, true);
		}

		/* file:{unlink} or dir:{DIR__RMDIR}, if necessary */
		tclass = sepgsql_file_class(newpath, false);
		if (tclass == SEPG_CLASS_DIR)
			sepgsql_file_common(newpath, tclass, SEPG_DIR__RMDIR, true);
		else if (tclass > 0)
			sepgsql_file_common(newpath, tclass, SEPG_FILE__UNLINK, true);
	}
#endif
}

void
sepgsql_dir_list(const char *dirname)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		uint16	tclass = sepgsql_file_class(dirname, true);

		if (tclass != SEPG_CLASS_DIR)
			ereport(ERROR,
					(errcode(ERRCODE_WRONG_OBJECT_TYPE),
					 errmsg("\"%s\" is not directory", dirname)));

		sepgsql_file_common(dirname, tclass, SEPG_DIR__READ, true);
	}
#endif
}
