/*
 * label.c
 *    SE-PostgreSQL security label management
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "access/relscan.h"
#include "access/xact.h"
#include "catalog/pg_attribute.h"
#include "catalog/pg_class.h"
#include "catalog/pg_database.h"
#include "catalog/pg_largeobject_metadata.h"
#include "catalog/pg_namespace.h"
#include "catalog/pg_proc.h"
#include "catalog/pg_seclabel.h"
#include "catalog/pg_type.h"
#include "commands/dbcommands.h"
#include "miscadmin.h"
#include "sepgsql/sepgsql.h"
#include "utils/fmgroids.h"
#include "utils/lsyscache.h"
#include "utils/rel.h"
#include "utils/syscache.h"
#include "utils/tqual.h"
#include <fnmatch.h>	// for workaround hack
#include <selinux/selinux.h>
#include <selinux/label.h>

/*
 * guc parameter to turn on/off mcstrans
 */
bool sepgsql_mcstrans;

sepgsql_sid_t
sepgsql_move_secid(Oid dst_relid, sepgsql_sid_t ssid)
{
	char		   *label;
	sepgsql_sid_t	dsid = { .relid = dst_relid,
							 .secid = InvalidOid };

	label = seclabelRawOutput(ssid.relid, ssid.secid);
	if (label)
	{
		dsid.secid = seclabelRawInput(dsid.relid, label);

		pfree(label);
	}
	return dsid;
}

static sepgsql_sid_t
get_default_secid_with_database(Oid relOid, Oid databaseOid, uint16 tclass)
{
	HeapTuple		tuple;
	sepgsql_sid_t	datsid;

	tuple = SearchSysCache1(DATABASEOID,
							ObjectIdGetDatum(databaseOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for database: %u", databaseOid);

	datsid.relid = DatabaseRelationId;
	datsid.secid = HeapTupleGetSecid(tuple);

	ReleaseSysCache(tuple);

	return sepgsql_client_create_secid(datsid, tclass, relOid);
}

static sepgsql_sid_t
get_default_secid_with_schema(Oid relOid, Oid namespaceOid, uint16 tclass)
{
	HeapTuple		tuple;
	sepgsql_sid_t	nspsid;

	tuple = SearchSysCache1(NAMESPACEOID,
							ObjectIdGetDatum(namespaceOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for namespace: %u", namespaceOid);

	nspsid.relid = NamespaceRelationId;
	nspsid.secid = HeapTupleGetSecid(tuple);

	ReleaseSysCache(tuple);

	return sepgsql_client_create_secid(nspsid, tclass, relOid);
}

static sepgsql_sid_t
get_default_secid_with_table(Oid relOid, Oid tableOid, uint16 tclass)
{
	HeapTuple		tuple;
	sepgsql_sid_t	tblsid;

	tuple = SearchSysCache1(RELOID,
							ObjectIdGetDatum(tableOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for relation: %u", tableOid);

	tblsid.relid = RelationRelationId;
	tblsid.secid = HeapTupleGetSecid(tuple);

	ReleaseSysCache(tuple);

	return sepgsql_client_create_secid(tblsid, tclass, relOid);
}

sepgsql_sid_t
sepgsql_get_default_database_secid(Oid templateOid)
{
	return get_default_secid_with_database(DatabaseRelationId,
										   templateOid,
										   SEPG_CLASS_DB_DATABASE);
}

sepgsql_sid_t
sepgsql_get_default_schema_secid(Oid databaseOid)
{
	return get_default_secid_with_database(NamespaceRelationId,
										   databaseOid,
										   SEPG_CLASS_DB_SCHEMA);
}

sepgsql_sid_t
sepgsql_get_default_table_secid(Oid namespaceOid)
{
	return get_default_secid_with_schema(RelationRelationId,
										 namespaceOid,
										 SEPG_CLASS_DB_TABLE);
}

sepgsql_sid_t
sepgsql_get_default_sequence_secid(Oid namespaceOid)
{
	return get_default_secid_with_schema(RelationRelationId,
										 namespaceOid,
										 SEPG_CLASS_DB_SEQUENCE);
}

sepgsql_sid_t
sepgsql_get_default_view_secid(Oid namespaceOid)
{
	return get_default_secid_with_schema(RelationRelationId,
										 namespaceOid,
										 SEPG_CLASS_DB_VIEW);
}

sepgsql_sid_t
sepgsql_get_default_proc_secid(Oid namespaceOid)
{
	return get_default_secid_with_schema(ProcedureRelationId,
										 namespaceOid,
										 SEPG_CLASS_DB_PROCEDURE);
}

sepgsql_sid_t
sepgsql_get_default_column_secid(Oid tableOid)
{
	return get_default_secid_with_table(AttributeRelationId,
										tableOid,
										SEPG_CLASS_DB_COLUMN);
}

sepgsql_sid_t
sepgsql_get_default_tuple_secid(Oid tableOid)
{
	return get_default_secid_with_table(tableOid,
										tableOid,
										SEPG_CLASS_DB_TUPLE);
}

sepgsql_sid_t
sepgsql_get_default_blob_secid(Oid databaseOid)
{
	return get_default_secid_with_database(LargeObjectMetadataRelationId,
										   databaseOid,
										   SEPG_CLASS_DB_BLOB);
}

Oid
sepgsql_get_default_secid(Relation rel, HeapTuple tuple)
{
	Oid				namespaceId;
	sepgsql_sid_t	nsid
		= { .relid = RelationGetRelid(rel), .secid = InvalidOid };

	switch (RelationGetRelid(rel))
	{
		case DatabaseRelationId:
		case RelationRelationId:
		case AttributeRelationId:
			elog(WARNING, "Bug? a new tuple without security id on \"%s\"",
				 RelationGetRelationName(rel));
			break;

		case NamespaceRelationId:
			nsid = sepgsql_get_default_schema_secid(MyDatabaseId);
			break;

		case ProcedureRelationId:
			namespaceId = ((Form_pg_proc) GETSTRUCT(tuple))->pronamespace;
			nsid = sepgsql_get_default_proc_secid(namespaceId);
			break;

		case LargeObjectMetadataRelationId:
			nsid = sepgsql_get_default_blob_secid(MyDatabaseId);
			break;

		default:
			nsid = sepgsql_get_default_tuple_secid(RelationGetRelid(rel));
			break;
	}
	return nsid.secid;
}

static Oid
sepgsql_initial_labeling_lookup(struct selabel_handle *sehnd,
								Oid relationId, HeapTuple tuple)
{
	Form_pg_database	datForm;
	Form_pg_namespace	nspForm;
	Form_pg_class		clsForm;
	Form_pg_attribute	attForm;
	Form_pg_proc		proForm;
	char				namebuf[NAMEDATALEN * 3 + 10];
	Oid					namespaceId;
	Oid					securityId = InvalidOid;
	int					type = 0;
	security_context_t	context;

	switch (relationId)
	{
		case DatabaseRelationId:
			datForm = (Form_pg_database) GETSTRUCT(tuple);
			snprintf(namebuf, sizeof(namebuf), "%s",
					 NameStr(datForm->datname));
			type = SELABEL_DB_DATABASE;
			break;

		case NamespaceRelationId:
			nspForm = (Form_pg_namespace) GETSTRUCT(tuple);
			snprintf(namebuf, sizeof(namebuf), "%s.%s",
					 get_database_name(MyDatabaseId),
					 NameStr(nspForm->nspname));
			type = SELABEL_DB_SCHEMA;
			break;

		case RelationRelationId:
			clsForm = (Form_pg_class) GETSTRUCT(tuple);
			namespaceId = clsForm->relnamespace;
			switch (clsForm->relkind)
			{
				case RELKIND_RELATION:
					snprintf(namebuf, sizeof(namebuf), "%s.%s.%s",
							 get_database_name(MyDatabaseId),
							 get_namespace_name(namespaceId),
							 NameStr(clsForm->relname));
					type = SELABEL_DB_TABLE;
					break;

				case RELKIND_SEQUENCE:
					snprintf(namebuf, sizeof(namebuf), "%s.%s.%s",
							 get_database_name(MyDatabaseId),
							 get_namespace_name(namespaceId),
							 NameStr(clsForm->relname));
					type = SELABEL_DB_SEQUENCE;
					break;

				case RELKIND_VIEW:
					snprintf(namebuf, sizeof(namebuf), "%s.%s.%s",
							 get_database_name(MyDatabaseId),
							 get_namespace_name(namespaceId),
							 NameStr(clsForm->relname));
					type = SELABEL_DB_VIEW;
					break;

				case RELKIND_INDEX: {
					HeapTuple	tbltup;
					HeapTuple	indtup;
					Oid			tableId;

					indtup = SearchSysCache1(INDEXRELID,
											 ObjectIdGetDatum(HeapTupleGetOid(tuple)));
					if (!HeapTupleIsValid(indtup))
						elog(ERROR, "cache lookup failed for index %u",
							 HeapTupleGetOid(tuple));

					tableId = ((Form_pg_index) GETSTRUCT(indtup))->indrelid;
					tbltup = SearchSysCache1(RELOID, ObjectIdGetDatum(tableId));
					if (!HeapTupleIsValid(tbltup))
						elog(ERROR, "cache lookup failed for relation %u", tableId);

					securityId = sepgsql_initial_labeling_lookup(sehnd, RelationRelationId, tbltup);

					ReleaseSysCache(tbltup);
					ReleaseSysCache(indtup);

					return securityId;
				}
				case RELKIND_TOASTVALUE: {
					HeapTuple	tbltup;
					Oid			tableId;

					/*
					 * XXX - we assume all the toast relation's name is
					 * "pg_toast_%u", and the "%u" shall be replaced by OID
					 * of the relation which owns the toast relation
					 */
					tableId = strtoul(NameStr(clsForm->relname) + 9, NULL, 10);

					tbltup = SearchSysCache1(RELOID,
											 ObjectIdGetDatum(tableId));
					if (!HeapTupleIsValid(tbltup))
						elog(ERROR, "cache lookup failed for relation %u", tableId);

					securityId = sepgsql_initial_labeling_lookup(sehnd, RelationRelationId, tbltup);

					ReleaseSysCache(tbltup);

					return securityId;
				}
				case RELKIND_COMPOSITE_TYPE: {
					HeapTuple	typtup;

					typtup = SearchSysCache1(TYPEOID,
											 ObjectIdGetDatum(clsForm->reltype));
					if (!HeapTupleIsValid(typtup))
						elog(ERROR, "cache lookup failed for type %u", clsForm->reltype);

					securityId = sepgsql_initial_labeling_lookup(sehnd, TypeRelationId, typtup);

					ReleaseSysCache(typtup);

					return securityId;
				}
				default:
					elog(ERROR, "unexpected relkind %c of \"%s\"",
						 clsForm->relkind, NameStr(clsForm->relname));
					break;
			}
			break;

		case AttributeRelationId:
			attForm = (Form_pg_attribute) GETSTRUCT(tuple);
			if (get_rel_relkind(attForm->attrelid) == RELKIND_RELATION)
			{
				namespaceId = get_rel_namespace(attForm->attrelid);
				snprintf(namebuf, sizeof(namebuf), "%s.%s.%s.%s",
						 get_database_name(MyDatabaseId),
						 get_namespace_name(namespaceId),
						 get_rel_name(attForm->attrelid),
						 NameStr(attForm->attname));
				type = SELABEL_DB_COLUMN;
			}
			else
			{
				HeapTuple	classTup;

				classTup = SearchSysCache1(RELOID, ObjectIdGetDatum(attForm->attrelid));
				if (!HeapTupleIsValid(classTup))
					elog(ERROR, "cache lookup failed for relation %u", attForm->attrelid);

				securityId = sepgsql_initial_labeling_lookup(sehnd, RelationRelationId, classTup);

				ReleaseSysCache(classTup);

				return securityId;
			}
			break;

		case ProcedureRelationId:
			proForm = (Form_pg_proc) GETSTRUCT(tuple);
			namespaceId = proForm->pronamespace;
			snprintf(namebuf, sizeof(namebuf), "%s.%s.%s",
					 get_database_name(MyDatabaseId),
					 get_namespace_name(namespaceId),
					 NameStr(proForm->proname));
			type = SELABEL_DB_PROCEDURE;
			break;

		case LargeObjectMetadataRelationId:
			snprintf(namebuf, sizeof(namebuf), "%s.%u",
					 get_database_name(MyDatabaseId),
					 HeapTupleGetOid(tuple));
			type = SELABEL_DB_BLOB;
			break;

		default:
			namespaceId = get_rel_namespace(relationId);
			snprintf(namebuf, sizeof(namebuf), "%s.%s.%s",
					 get_database_name(MyDatabaseId),
					 get_namespace_name(namespaceId),
					 get_rel_name(relationId));
			type = SELABEL_DB_TUPLE;
			break;
	}

	if (selabel_lookup_raw(sehnd, &context, namebuf, type) == 0)
	{
		PG_TRY();
		{
			securityId = seclabelTransInput(relationId, context);
		}
		PG_CATCH();
		{
			freecon(context);
			PG_RE_THROW();
		}
		PG_END_TRY();
		freecon(context);
	}
	else if (errno == ENOENT)
		elog(WARNING,
			 "SELinux: no initial label assigned for %s (type=%d), skipping",
			 namebuf, type);
	else
		elog(ERROR,
			 "SELinux: could not determine initial security label for %s (type=%d): %m",
			 namebuf, type);

	return securityId;
}

void
sepgsql_initial_labeling(void)
{
	Relation		classRel;
	SysScanDesc		classScan;
	ScanKeyData		classSkey;
	HeapTuple		classTup;
	struct selabel_handle  *sehnd;

	Assert(IsBootstrapProcessingMode());

	StartTransactionCommand();

	sehnd = selabel_open(SELABEL_CTX_DB, NULL, 0);
	if (!sehnd)
		elog(ERROR, "failed to open initial selinux label");

	classRel = heap_open(RelationRelationId, AccessShareLock);

	ScanKeyInit(&classSkey,
				Anum_pg_class_relhassecids,
				BTEqualStrategyNumber, F_BOOLEQ,
				BoolGetDatum(true));

	classScan = systable_beginscan(classRel, InvalidOid, false,
								   SnapshotNow, 1, &classSkey);

	while (HeapTupleIsValid(classTup = systable_getnext(classScan)))
	{
		Relation		rel;
		HeapScanDesc	scan;
		HeapTuple		oldtup;
		HeapTuple		newtup;
		Oid				relOid = HeapTupleGetOid(classTup);

		Assert(((Form_pg_class) GETSTRUCT(classTup))->relhassecids);

		rel = heap_open(relOid, RowExclusiveLock);

		scan = heap_beginscan(rel, SnapshotNow, 0, NULL);

		while (HeapTupleIsValid(oldtup = heap_getnext(scan, ForwardScanDirection)))
		{
			Oid		securityId = sepgsql_initial_labeling_lookup(sehnd, relOid, oldtup);

			/*
			 * inplace-updating
			 */
			newtup = heap_copytuple(oldtup);

			HeapTupleSetSecid(newtup, securityId);

			heap_inplace_update(rel, newtup);

			heap_freetuple(newtup);
		}

		heap_endscan(scan);
		
		heap_close(rel, RowExclusiveLock);
	}
	systable_endscan(classScan);

	heap_close(classRel, AccessShareLock);

	selabel_close(sehnd);

	CommitTransactionCommand();
}

char *
sepgsql_mcstrans_in(char *trans_label)
{
	security_context_t	raw_label;
	security_context_t	result;

	if (!sepgsql_mcstrans)
		return trans_label;

	if (selinux_trans_to_raw_context(trans_label, &raw_label) < 0)
		ereport(ERROR,
				(errcode(ERRCODE_INTERNAL_ERROR),
				 errmsg("SELinux: unable to translate \"%s\"", trans_label)));
	PG_TRY();
	{
		result = pstrdup(raw_label);
	}
	PG_CATCH();
	{
		freecon(raw_label);
		PG_RE_THROW();
	}
	PG_END_TRY();
	freecon(raw_label);

	return result;
}

char *
sepgsql_mcstrans_out(char *raw_label)
{
	security_context_t	trans_label;
	security_context_t	result;

	if (!sepgsql_mcstrans)
		return raw_label;

	if (selinux_raw_to_trans_context(raw_label, &trans_label) < 0)
		ereport(ERROR,
				(errcode(ERRCODE_INTERNAL_ERROR),
				 errmsg("SELinux: unable to translate \"%s\"", raw_label)));
	PG_TRY();
	{
		result = pstrdup(trans_label);
	}
	PG_CATCH();
	{
		freecon(trans_label);
		PG_RE_THROW();
	}
	PG_END_TRY();
	freecon(trans_label);

	return result;
}

char *
sepgsql_rawlabel_in(char *label)
{
	if (!label || security_check_context_raw(label) < 0)
		ereport(ERROR,
				(errcode(ERRCODE_INVALID_NAME),
				 errmsg("invalid security context \"%s\"", label)));
	return label;
}

char *
sepgsql_rawlabel_out(char *label)
{
	if (!label || security_check_context_raw(label) < 0)
	{
		security_context_t	unlabeled_label;

		if (security_get_initial_context_raw("unlabeled",
											 &unlabeled_label) < 0)
			ereport(ERROR,
					(errcode(ERRCODE_INTERNAL_ERROR),
					 errmsg("unable to get unlabeled security context")));
		PG_TRY();
		{
			label = pstrdup(unlabeled_label);
		}
		PG_CATCH();
		{
			freecon(unlabeled_label);
			PG_RE_THROW();
		}
		PG_END_TRY();
		freecon(unlabeled_label);
	}
	return label;
}
