/*
 * operator.c 
 *
 * SELinux hooks related to operators
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "catalog/pg_operator.h"
#include "catalog/pg_opclass.h"
#include "catalog/pg_opfamily.h"
#include "catalog/pg_seclabel.h"
#include "sepgsql/hooks.h"
#include "sepgsql/sepgsql.h"
#include "utils/builtins.h"
#include "utils/lsyscache.h"
#include "utils/syscache.h"

#ifdef HAVE_SELINUX
bool
sepgsql_operator_common(Oid operOid, uint32 required, bool abort)
{
	sepgsql_sid_t	tsid;
	char		   *auname;
	bool			retval;

	tsid.relid = OperatorRelationId;
	tsid.secid = GetSysCacheSecid1(OPEROID, ObjectIdGetDatum(operOid));

	auname = format_operator(operOid);

	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_TUPLE,
								  required,
								  auname,
								  abort);
	pfree(auname);

	return retval;
}

bool
sepgsql_opclass_common(Oid opcOid, uint32 required, bool abort)
{
	sepgsql_sid_t	tsid;
	HeapTuple		tuple;
	const char	   *auname;
	bool			retval;

	tuple = SearchSysCache1(CLAOID, ObjectIdGetDatum(opcOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for operator class %u", opcOid);

	tsid.relid = OperatorClassRelationId;
	tsid.secid = HeapTupleGetSecid(tuple);

	auname = NameStr(((Form_pg_opclass) GETSTRUCT(tuple))->opcname);

	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_TUPLE,
								  required,
								  auname,
								  abort);
	ReleaseSysCache(tuple);

	return retval;
}

bool
sepgsql_opfamily_common(Oid opfOid, uint32 required, bool abort)
{
	sepgsql_sid_t	tsid;
	HeapTuple		tuple;
	const char	   *auname;
	bool			retval;

	tuple = SearchSysCache1(OPFAMILYOID, ObjectIdGetDatum(opfOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for operator family %u", opfOid);

	tsid.relid = OperatorFamilyRelationId;
	tsid.secid = HeapTupleGetSecid(tuple);

	auname = NameStr(((Form_pg_opfamily) GETSTRUCT(tuple))->opfname);

	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_TUPLE,
								  required,
								  auname,
								  abort);
	ReleaseSysCache(tuple);

	return retval;
}

static Oid
get_operator_namespace(Oid operOid)
{
    HeapTuple	tuple;
	Oid			namespaceId;

	tuple = SearchSysCache1(OPEROID, ObjectIdGetDatum(operOid));
	if (HeapTupleIsValid(tuple))
	{
		namespaceId = ((Form_pg_operator) GETSTRUCT(tuple))->oprnamespace;

		ReleaseSysCache(tuple);
	}
	return namespaceId;
}

static Oid
get_opclass_namespace(Oid opcOid)
{
	HeapTuple	tuple;
	Oid			namespaceId = InvalidOid;

	tuple = SearchSysCache1(CLAOID, ObjectIdGetDatum(opcOid));
	if (HeapTupleIsValid(tuple))
	{
		namespaceId = ((Form_pg_opclass) GETSTRUCT(tuple))->opcnamespace;

		ReleaseSysCache(tuple);
	}
	return namespaceId;
}

static Oid
get_opfamily_namespace(Oid opfOid)
{
	HeapTuple	tuple;
	Oid			namespaceId = InvalidOid;

	tuple = SearchSysCache1(OPFAMILYOID, ObjectIdGetDatum(opfOid));
	if (HeapTupleIsValid(tuple))
	{
		namespaceId = ((Form_pg_opfamily) GETSTRUCT(tuple))->opfnamespace;

		ReleaseSysCache(tuple);
	}
	return namespaceId;
}

#endif

Oid
sepgsql_operator_create(const char *operName, Oid replaced, Oid namespaceId,
						Oid codeFunc, Oid restrictFunc, Oid joinFunc,
						Oid commutatorOp, Oid negatorOp)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;
		uint32			required;

		/* db_schema:{add_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__ADD_NAME, true);

		/* db_tuple:{insert} */
		if (OidIsValid(replaced))
		{
			nsid.relid = OperatorRelationId;
			nsid.secid = GetSysCacheSecid1(OPEROID,
										   ObjectIdGetDatum(replaced));
			required = SEPG_DB_TUPLE__UPDATE;
		}
		else
		{
			nsid = sepgsql_get_default_tuple_secid(OperatorRelationId);
			required = SEPG_DB_TUPLE__INSERT;
		}
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_TUPLE,
							 required,
							 operName,
							 true);

		/* db_procedure:{install} */
		if (OidIsValid(codeFunc))
			sepgsql_proc_common(codeFunc, SEPG_DB_PROCEDURE__INSTALL, true);
		if (OidIsValid(restrictFunc))
			sepgsql_proc_common(restrictFunc, SEPG_DB_PROCEDURE__INSTALL, true);
		if (OidIsValid(joinFunc))
			sepgsql_proc_common(joinFunc, SEPG_DB_PROCEDURE__INSTALL, true);

		/*
		 * XXX - we should check anything on he commutatorOp/negatorOp
		 */

		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_operator_alter(Oid operOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_operator_common(operOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

Oid
sepgsql_operator_relabel(Oid operOid, char *newLabel)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;

		nsid.relid = OperatorRelationId;
		nsid.secid = seclabelTransInput(OperatorRelationId, newLabel);

		/* db_tuple:{update relabelfrom} */
		sepgsql_operator_common(operOid,
								SEPG_DB_TUPLE__UPDATE |
								SEPG_DB_TUPLE__RELABELFROM,
								true);

		/* db_tuple:{relabelto} */
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_TUPLE,
							 SEPG_DB_TUPLE__RELABELTO,
							 format_operator(operOid),
							 true);
		return nsid.secid;
	}
#endif
	ereport(ERROR,
			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
			 errmsg("SE-PostgreSQL is not available")));
	return InvalidOid;
}

void
sepgsql_operator_drop(Oid operOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		namespaceId = get_operator_namespace(operOid);

		/* db_namespace:{remove_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_tuple:{delete} */
		sepgsql_operator_common(operOid, SEPG_DB_TUPLE__DELETE, true);
	}
#endif
}

void
sepgsql_operator_comment(Oid operOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{update} */
		sepgsql_operator_common(operOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

Oid
sepgsql_opclass_create(const char *opcName, Oid namespaceId,
                Oid typeOid, Oid opfamilyOid, Oid storageOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;

		/* db_schema:{add_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_TUPLE__UPDATE, true);

		/* db_tuple:{insert} */
		nsid = sepgsql_get_default_tuple_secid(OperatorRelationId);
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_TUPLE,
							 SEPG_DB_TUPLE__INSERT,
							 opcName,
							 true);
		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_opclass_alter(Oid opcOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{update} */
		sepgsql_opclass_common(opcOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_opclass_alter_rename(Oid opcOid, const char *newName)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		namespaceId = get_opclass_namespace(opcOid);

		/* db_schema:{add_name remove_name} */
		sepgsql_schema_common(namespaceId,
							  SEPG_DB_SCHEMA__ADD_NAME |
							  SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_tuple:{update} */
		sepgsql_opclass_common(opcOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_opclass_drop(Oid opcOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		namespaceId = get_opclass_namespace(opcOid);

		/* db_schema:{remove_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_tuple:{drop} */
		sepgsql_opclass_common(opcOid, SEPG_DB_TUPLE__DELETE, true);
	}
#endif
}

void
sepgsql_opclass_comment(Oid opcOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{update} */
		sepgsql_opclass_common(opcOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

Oid
sepgsql_opfamily_create(const char *opfName, Oid namespaceId, Oid amOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;

		/* db_schema:{add_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__ADD_NAME, true);

		/* db_tuple:{insert} */
		nsid = sepgsql_get_default_tuple_secid(OperatorFamilyRelationId);
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_TUPLE,
							 SEPG_DB_TUPLE__INSERT,
							 opfName,
							 true);
		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_opfamily_alter(Oid opfOid, bool isDrop, Oid amOid,
					   List *operators, List *procedures)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{update} */
		sepgsql_opfamily_common(opfOid, SEPG_DB_TUPLE__UPDATE, true);

		/* XXX - to do we should install checks? */
	}
#endif
}

void
sepgsql_opfamily_alter_rename(Oid opfOid, const char *newName)
{

#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		namespaceId = get_opfamily_namespace(opfOid);

		/* db_schema:{add_name remove_name} */
		sepgsql_schema_common(namespaceId,
							  SEPG_DB_SCHEMA__ADD_NAME |
							  SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_tuple:{update} */
		sepgsql_opfamily_common(opfOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_opfamily_alter_owner(Oid opfOid, Oid newOwner)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{update} */
		sepgsql_opfamily_common(opfOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_opfamily_drop(Oid opfOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		namespaceId = get_opfamily_namespace(opfOid);

		/* db_schema:{remove_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_tuple:{delete} */
		sepgsql_opfamily_common(opfOid, SEPG_DB_TUPLE__DELETE, true);
	}
#endif
}

void
sepgsql_opfamily_comment(Oid opfOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{update} */
		sepgsql_opfamily_common(opfOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}
