/*
 * proc.c
 *
 * SELinux hooks related to procedures
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "catalog/pg_aggregate.h"
#include "catalog/pg_language.h"
#include "catalog/pg_namespace.h"
#include "catalog/pg_proc.h"
#include "catalog/pg_seclabel.h"
#include "sepgsql/hooks.h"
#include "sepgsql/sepgsql.h"
#include "utils/builtins.h"
#include "utils/syscache.h"
#include "utils/lsyscache.h"

#ifdef HAVE_SELINUX
bool
sepgsql_proc_common(Oid procOid, uint32 required, bool abort)
{
	sepgsql_sid_t	tsid;
	char		   *auname;
	bool			retval;

	tsid.relid = ProcedureRelationId;
	tsid.secid = GetSysCacheSecid1(PROCOID,
								   ObjectIdGetDatum(procOid));

	auname = format_procedure(procOid);

	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_PROCEDURE,
								  required,
								  auname,
								  abort);
	pfree(auname);

	return retval;
}
#endif

Oid
sepgsql_proc_create(const char *proName, Oid replaced,
					Oid namespaceId, Oid languageId)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		HeapTuple		tuple;
		sepgsql_sid_t	nsid;
		char		   *scontext;
		char		   *tcontext;
		uint32			required;

		/* db_schema:{add_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__ADD_NAME, true);

		/* db_procedure:{create or setattr} */
		if (OidIsValid(replaced))
		{
			nsid.relid = ProcedureRelationId;
			nsid.secid = GetSysCacheSecid1(PROCOID,
										   ObjectIdGetDatum(replaced));
			required = SEPG_DB_PROCEDURE__SETATTR;
		}
		else
		{
			nsid = sepgsql_get_default_proc_secid(namespaceId);
			required = SEPG_DB_PROCEDURE__CREATE;
		}
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_PROCEDURE,
							 required,
							 proName,
							 true);

		/* db_language:{implemente} */
		tuple = SearchSysCache1(LANGOID, ObjectIdGetDatum(languageId));
		if (!HeapTupleIsValid(tuple))
			elog(ERROR, "cache lookup failed for language %u", languageId);

		scontext = seclabelRawOutput(LanguageRelationId,
									 HeapTupleGetSecid(tuple));
		tcontext = seclabelRawOutput(nsid.relid, nsid.secid);

		sepgsql_compute_perms(scontext, tcontext,
							  SEPG_CLASS_DB_LANGUAGE,
							  SEPG_DB_LANGUAGE__IMPLEMENTE,
							  proName, true);

		ReleaseSysCache(tuple);

		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_proc_alter(Oid procOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_proc_common(procOid, SEPG_DB_PROCEDURE__SETATTR, true);
	}
#endif
}

void
sepgsql_proc_alter_rename(Oid procOid, const char *newName)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		namespaceId = get_func_namespace(procOid);

		/* db_schema:{remove_name add_name} */
		sepgsql_schema_common(namespaceId,
							  SEPG_DB_SCHEMA__REMOVE_NAME |
							  SEPG_DB_SCHEMA__ADD_NAME, true);

		/* db_procedure:{setattr} */
		sepgsql_proc_common(procOid, SEPG_DB_PROCEDURE__SETATTR, true);
	}
#endif
}

void
sepgsql_proc_alter_schema(Oid procOid, Oid newSchema)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		namespaceId = get_func_namespace(procOid);

		/* db_schema:{remove_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_schema:{add_name} */
		sepgsql_schema_common(newSchema, SEPG_DB_SCHEMA__ADD_NAME, true);

		/* db_procedure:{setattr} */
		sepgsql_proc_common(procOid, SEPG_DB_PROCEDURE__SETATTR, true);
	}
#endif
}

Oid
sepgsql_proc_relabel(Oid procOid, char *new_label)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	tsid;

		tsid.relid = ProcedureRelationId;
		tsid.secid = seclabelTransInput(tsid.relid, new_label);

		/* db_procedure:{setattr relabelfrom} */
		sepgsql_proc_common(procOid,
							SEPG_DB_PROCEDURE__SETATTR |
							SEPG_DB_PROCEDURE__RELABELFROM, true);

		/* db_procedure:{relabelto} */
		sepgsql_client_perms(tsid,
							 SEPG_CLASS_DB_PROCEDURE,
							 SEPG_DB_PROCEDURE__RELABELTO,
							 format_procedure(procOid),
							 true);
		return tsid.secid;
	}
#endif
	ereport(ERROR,
			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
			 errmsg("SE-PostgreSQL is not available")));
	return InvalidOid;
}

void
sepgsql_proc_drop(Oid procOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		namespaceId = get_func_namespace(procOid);

		/* db_schema:{remove_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_procedure:{proc} */
		sepgsql_proc_common(procOid, SEPG_DB_PROCEDURE__DROP, true);
	}
#endif
}

void
sepgsql_proc_grant(Oid procOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_proc_common(procOid, SEPG_DB_PROCEDURE__SETATTR, true);
	}
#endif
}

void
sepgsql_proc_comment(Oid procOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_proc_common(procOid, SEPG_DB_PROCEDURE__SETATTR, true);
	}
#endif
}

void
sepgsql_proc_execute(Oid procOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_proc_common(procOid, SEPG_DB_PROCEDURE__EXECUTE, true);
	}
#endif
}

bool
sepgsql_proc_be_inlined(HeapTuple protup)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	tsid;
		char		   *old_label;
		char		   *new_label;
		const char	   *auname
			= NameStr(((Form_pg_proc) GETSTRUCT(protup))->proname);

		tsid.relid = ProcedureRelationId;
		tsid.secid = HeapTupleGetSecid(protup);

		if (!sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_PROCEDURE,
								  SEPG_DB_PROCEDURE__EXECUTE,
								  auname, false))
			return false;

		old_label = sepgsql_get_client_label();
		new_label = sepgsql_client_create_label(tsid, SEPG_CLASS_PROCESS);
		if (strcmp(old_label, new_label) != 0)
			return false;

		return true;
	}
#endif
	return true;
}

char *
sepgsql_proc_domtrans(HeapTuple protup, MemoryContext mcxt)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	tsid;
		char		   *old_label = sepgsql_get_client_label();
		char		   *new_label;
		char		   *auname
			= NameStr(((Form_pg_proc) GETSTRUCT(protup))->proname);

		tsid.relid = ProcedureRelationId;
		tsid.secid = HeapTupleGetSecid(protup);

		new_label = sepgsql_client_create_label(tsid, SEPG_CLASS_PROCESS);

		if (strcmp(old_label, new_label) == 0)
			return NULL;

		/* db_procedure:{entrypoint} */
		sepgsql_client_perms(tsid,
							 SEPG_CLASS_DB_PROCEDURE,
							 SEPG_DB_PROCEDURE__ENTRYPOINT,
							 auname,
							 true);

		/* db_process:{transition} */
		sepgsql_compute_perms(old_label,
							  new_label,
							  SEPG_CLASS_PROCESS,
							  SEPG_PROCESS__TRANSITION,
							  NULL,
							  true);

		return MemoryContextStrdup(mcxt, new_label);
	}
#endif
	return NULL;
}

Oid
sepgsql_aggregate_create(const char *aggName, Oid namespaceId,
						 Oid transFunc, Oid finalFunc)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		securityId;

		/* check normal creation permission */
		securityId = sepgsql_proc_create(aggName, InvalidOid,
										 namespaceId, INTERNALlanguageId);

		/* db_procedure:{install} */
		if (OidIsValid(transFunc))
			sepgsql_proc_common(transFunc, SEPG_DB_PROCEDURE__INSTALL, true);

		/* db_procedure:{install} */
		if (OidIsValid(finalFunc))
			sepgsql_proc_common(finalFunc, SEPG_DB_PROCEDURE__INSTALL, true);

		return securityId;
	}
#endif
	return InvalidOid;
}

void
sepgsql_aggregate_execute(Oid aggOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Form_pg_aggregate	aggForm;
		HeapTuple			tuple;

		tuple = SearchSysCache1(AGGFNOID, ObjectIdGetDatum(aggOid));
		if (!HeapTupleIsValid(tuple))
			elog(ERROR, "cache lookup failed for aggregate %u", aggOid);

		aggForm = (Form_pg_aggregate) GETSTRUCT(tuple);

		/* pg_proc:{execute} */
		sepgsql_proc_common(aggOid, SEPG_DB_PROCEDURE__EXECUTE, true);

		/* pg_proc:{execute} */
		if (OidIsValid(aggForm->aggtransfn))
			sepgsql_proc_common(aggForm->aggtransfn,
								SEPG_DB_PROCEDURE__EXECUTE, true);

		/* pg_proc:{execute} */
		if (OidIsValid(aggForm->aggfinalfn))
			sepgsql_proc_common(aggForm->aggfinalfn,
								SEPG_DB_PROCEDURE__EXECUTE, true);

		ReleaseSysCache(tuple);
	}
#endif
}
