/*
 * relation.c
 *
 * SELinux hooks related to relation
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "access/sysattr.h"
#include "catalog/catalog.h"
#include "catalog/heap.h"
#include "catalog/pg_attribute.h"
#include "catalog/pg_class.h"
#include "catalog/pg_seclabel.h"
#include "catalog/pg_type.h"
#include "sepgsql/hooks.h"
#include "sepgsql/sepgsql.h"
#include "utils/lsyscache.h"
#include "utils/syscache.h"

#ifdef HAVE_SELINUX
bool
sepgsql_relation_common(Oid relOid, uint32 required, bool abort)
{
	Form_pg_class	classForm;
	HeapTuple		tuple;
	sepgsql_sid_t	tsid;
	uint16			tclass;
	bool			retval;

	tuple = SearchSysCache1(RELOID, ObjectIdGetDatum(relOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for relation %u", relOid);
	classForm = (Form_pg_class) GETSTRUCT(tuple);

	tsid.relid = RelationRelationId;
	tsid.secid = HeapTupleGetSecid(tuple);

	switch (classForm->relkind)
	{
		case RELKIND_SEQUENCE:
			tclass = SEPG_CLASS_DB_SEQUENCE;
			break;

		case RELKIND_VIEW:
			tclass = SEPG_CLASS_DB_VIEW;
			break;

		case RELKIND_COMPOSITE_TYPE:
			tclass = SEPG_CLASS_DB_TUPLE;
			break;

		default:
			tclass = SEPG_CLASS_DB_TABLE;
			break;
	}

	retval = sepgsql_client_perms(tsid,
								  tclass,
								  required,
								  NameStr(classForm->relname),
								  abort);
	ReleaseSysCache(tuple);

	return retval;
}

static Bitmapset *
fixup_whole_row_reference(Oid relOid, int natts, Bitmapset *columns)
{
	Bitmapset  *result;
	AttrNumber	attno;

	attno = InvalidAttrNumber - FirstLowInvalidHeapAttributeNumber;
	if (!bms_is_member(attno, columns))
		return columns;	/* no need to fixup */

	result = bms_copy(columns);
	result = bms_del_member(result, attno);

	for (attno=1; attno <= natts; attno++)
	{
		Form_pg_attribute	attForm;
		HeapTuple			atttup;

		atttup = SearchSysCache2(ATTNUM,
								 ObjectIdGetDatum(relOid),
								 Int16GetDatum(attno));
		if (!HeapTupleIsValid(atttup))
			continue;

		attForm = (Form_pg_attribute) GETSTRUCT(atttup);
		if (!attForm->attisdropped)
		{
			int		cindex = attno - FirstLowInvalidHeapAttributeNumber;
			result = bms_add_member(result, cindex);
		}
		ReleaseSysCache(atttup);
	}

	return result;
}
#endif

bool
sepgsql_relation_perms(Oid relOid, AclMode aclmask,
					   Bitmapset *selectedCols,
					   Bitmapset *modifiedCols, bool abort)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Form_pg_class	relForm;
		HeapTuple		tuple;
		sepgsql_sid_t	rsid;
		Bitmapset	   *selColsEx;
		Bitmapset	   *modColsEx;
		Bitmapset	   *columns;
		AttrNumber		nattrs;
		AttrNumber		attno;
		const char	   *auname;
		char			relkind;
		uint16			tclass = 0;
		uint32			required = 0;
		bool			rc = true;

		/*
		 * Hardwired policy:
		 * SE-PostgreSQL enforces clients cannot modify system catalogs
		 * and access toast values using DML statements in enforcing mode.
		 * Note that it performs in permissive mode during initdb phase.
		 */
		if (sepgsql_get_enforce())
		{
			if (IsSystemNamespace(get_rel_namespace(relOid)) &&
				(aclmask & (ACL_UPDATE | ACL_INSERT | ACL_DELETE)) != 0)
				ereport(ERROR,
						(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
						 errmsg("not allowed to modify system catalog \"%s\"",
								get_rel_name(relOid))));

			if (get_rel_relkind(relOid) == RELKIND_TOASTVALUE)
				ereport(ERROR,
						(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
						 errmsg("not allowed to access toast values \"%s\"",
								get_rel_name(relOid))));
		}

		/*
		 * check relation's permissions
		 */
		tuple = SearchSysCache1(RELOID,
								ObjectIdGetDatum(relOid));
		if (!HeapTupleIsValid(tuple))
			elog(ERROR, "cache lookup failed for relation %u", relOid);

		rsid.relid = RelationRelationId;
		rsid.secid = HeapTupleGetSecid(tuple);

		relForm = (Form_pg_class) GETSTRUCT(tuple);
		auname = NameStr(relForm->relname);
		relkind = relForm->relkind;
		nattrs = relForm->relnatts;

		switch (relkind)
		{
			case RELKIND_RELATION:
				tclass = SEPG_CLASS_DB_TABLE;

				if (aclmask & ACL_SELECT)
					required |= SEPG_DB_TABLE__SELECT;
				if (aclmask & ACL_INSERT)
					required |= SEPG_DB_TABLE__INSERT;
				if (aclmask & ACL_UPDATE)
					required |= (!modifiedCols
								 ? SEPG_DB_TABLE__LOCK
								 : SEPG_DB_TABLE__UPDATE);
				if (aclmask & ACL_DELETE)
					required |= SEPG_DB_TABLE__DELETE;
				break;

			case RELKIND_SEQUENCE:
				tclass = SEPG_CLASS_DB_SEQUENCE;
				if (aclmask & ACL_SELECT)
					required |= SEPG_DB_SEQUENCE__GET_VALUE;
				break;

			case RELKIND_VIEW:
				tclass = SEPG_CLASS_DB_VIEW;
				if (aclmask != 0)
					required |= SEPG_DB_VIEW__EXPAND;
				break;

			default:
				elog(ERROR, "Bug? unexpected relkind %c", relkind);
				return false;
		}

		if (required != 0)
			rc = sepgsql_client_perms(rsid, tclass, required, auname, abort);

		ReleaseSysCache(tuple);

		if (!rc || relkind != RELKIND_RELATION)
			return rc;

		/*
		 * Check column's permissions
		 */
		selColsEx = fixup_whole_row_reference(relOid, nattrs,
											  selectedCols);
		modColsEx = fixup_whole_row_reference(relOid, nattrs,
											  modifiedCols);
		columns = bms_union(selColsEx, modColsEx);

		while ((attno = bms_first_member(columns)) >= 0)
		{
			required = 0;

			if (bms_is_member(attno, selColsEx))
				required |= SEPG_DB_COLUMN__SELECT;
			if (bms_is_member(attno, modColsEx))
			{
				if (aclmask & ACL_UPDATE)
					required |= SEPG_DB_COLUMN__UPDATE;
				if (aclmask & ACL_INSERT)
					required |= SEPG_DB_COLUMN__INSERT;
			}
			if (required == 0)
				continue;

			attno += FirstLowInvalidHeapAttributeNumber;
			rc = sepgsql_attribute_common(relOid, attno, required, abort);
			if (!rc)
				break;
		}

		if (selColsEx != selectedCols)
			bms_free(selColsEx);
		if (modColsEx != modifiedCols)
			bms_free(modColsEx);
		bms_free(columns);

		return rc;
	}
#endif
	return true;
}

Oid *
sepgsql_relation_create(const char *relName,
						char relkind,
						TupleDesc tupDesc,
						Oid namespaceId,
						List *supOids,
						bool createAs)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	tsid;
		sepgsql_sid_t	csid;
		Oid			   *secLabels;
		uint16			tclass;
		uint32			perms;
		AttrNumber		index, attno, nitems;

		/*
		 * The secLabeld array stores security identifiers to be
		 * assigned on the new table and columns.
		 *
		 * secLabels[0] is security-id of the relation.
		 * secLabels[attnum - FirstLowInvalidHeapAttributeNumber]
		 *   is security-id of the columns.
		 */
		secLabels = seclabelMakeRelationDefaults(tupDesc, supOids);
		nitems = tupDesc->natts - FirstLowInvalidHeapAttributeNumber;

		switch (relkind)
		{
			case RELKIND_RELATION:
				if (!OidIsValid(secLabels[0]))
				{
					tsid = sepgsql_get_default_table_secid(namespaceId);
					secLabels[0] = tsid.secid;
				}
				tclass = SEPG_CLASS_DB_TABLE;
				perms = SEPG_DB_TABLE__CREATE;
				if (createAs)
					perms |= SEPG_DB_TABLE__INSERT;
				break;

			case RELKIND_SEQUENCE:
				if (!OidIsValid(secLabels[0]))
				{
					tsid = sepgsql_get_default_sequence_secid(namespaceId);
					secLabels[0] = tsid.secid;
				}
				tclass = SEPG_CLASS_DB_SEQUENCE;
				perms = SEPG_DB_SEQUENCE__CREATE;
				break;

			case RELKIND_VIEW:
				if (!OidIsValid(secLabels[0]))
				{
					tsid = sepgsql_get_default_view_secid(namespaceId);
					secLabels[0] = tsid.secid;
				}
				tclass = SEPG_CLASS_DB_VIEW;
				perms = SEPG_DB_VIEW__CREATE;
				break;

			case RELKIND_COMPOSITE_TYPE:
				if (!OidIsValid(secLabels[0]))
				{
					tsid = sepgsql_get_default_tuple_secid(TypeRelationId);
					secLabels[0] = seclabelMoveSecid(RelationRelationId,
													 TypeRelationId, tsid.secid);
				}
				tclass = SEPG_CLASS_DB_TUPLE;
				perms = SEPG_DB_TUPLE__INSERT;
				break;

			default:
				elog(ERROR, "Bug? unexpected relkind %c", relkind);
				return NULL;
		}
		/* db_schema:{add_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__ADD_NAME, true);

		/* db_table:{create} or others */
		tsid.relid = RelationRelationId;
		tsid.secid = secLabels[0];
		sepgsql_client_perms(tsid, tclass, perms, relName, true);

		/* no individual security-id except for RELKIND_RELATION */
		if (relkind != RELKIND_RELATION)
		{
			Oid	securityId = seclabelMoveSecid(AttributeRelationId,
											   RelationRelationId, secLabels[0]);

			for (index = 1; index < nitems; index++)
				secLabels[index] = securityId;

			return secLabels;
		}

		/*
		 * security context of the columns
		 */
		for (index = 1; index < nitems; index++)
		{
			Form_pg_attribute	attForm;
			char		auname[NAMEDATALEN * 2 + 10];

			attno = index + FirstLowInvalidHeapAttributeNumber;

			/* skip unnecessary system columns */
			if ((attno == ObjectIdAttributeNumber && !tupDesc->tdhasoid) ||
				(attno == SecurityLabelAttributeNumber && !tupDesc->tdhassecid))
				continue;

			if (!OidIsValid(secLabels[index]))
			{
				csid = sepgsql_client_create_secid(tsid,
												   SEPG_CLASS_DB_COLUMN,
												   AttributeRelationId);
				secLabels[index] = csid.secid;
			}

			if (attno < 0)
				attForm = SystemAttributeDefinition(attno,
													tupDesc->tdhasoid,
													tupDesc->tdhassecid);
			else
				attForm = tupDesc->attrs[attno];

			/* db_column:{create (insert)} permission */
			csid.relid = AttributeRelationId;
			csid.secid = secLabels[index];

			perms = SEPG_DB_COLUMN__CREATE;
			if (createAs && attno >= 0)
				perms |= SEPG_DB_COLUMN__INSERT;

			snprintf(auname, sizeof(auname), "%s.%s",
					 relName, NameStr(attForm->attname));

			sepgsql_client_perms(csid,
								 SEPG_CLASS_DB_COLUMN,
								 perms,
								 auname,
								 true);
		}
		return secLabels;
	}
#endif
	return NULL;
}

void
sepgsql_relation_alter(Oid relationOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__SETATTR, true);
	}
#endif
}

void
sepgsql_relation_alter_schema(Oid relationOid, Oid newSchema)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{

		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__SETATTR, true);
	}
#endif
}

void
sepgsql_relation_alter_rename(Oid relationOid, const char *newName)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{

		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__SETATTR, true);
	}
#endif
}

void
sepgsql_relation_alter_inherit(Oid childOid, Oid parentOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_relation_common(childOid,
								SEPG_DB_TABLE__SETATTR, true);
	}
#endif
}

Oid
sepgsql_relation_relabel(Oid relationOid, char *new_label)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;
		char		   *auname;
		char			relkind;
		uint16			tclass = 0;

		nsid.relid = RelationRelationId;
		nsid.secid = seclabelTransInput(nsid.relid, new_label);

		auname = get_rel_name(relationOid);
		relkind = get_rel_relkind(relationOid);

		switch (relkind)
		{
			case RELKIND_RELATION:
				tclass = SEPG_CLASS_DB_TABLE;
				break;

			case RELKIND_SEQUENCE:
				tclass = SEPG_CLASS_DB_SEQUENCE;
				break;

			case RELKIND_VIEW:
				tclass = SEPG_CLASS_DB_VIEW;
				break;

			case RELKIND_COMPOSITE_TYPE:
				tclass = SEPG_CLASS_DB_TUPLE;
				break;

			default:
				elog(ERROR, "unexpected relkind %c", relkind);
				break;;
		}
		/* db_xxx:{setattr relabelfrom} */
		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__SETATTR |
								SEPG_DB_TABLE__RELABELFROM, true);

		/* db_xxx:{relabelto} */
		sepgsql_client_perms(nsid,
							 tclass,
							 SEPG_DB_TABLE__RELABELTO,
							 auname, true);
		pfree(auname);

	return nsid.secid;
	}
#endif
	ereport(ERROR,
			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
			 errmsg("SE-PostgreSQL is not available")));
	return InvalidOid;
}

void
sepgsql_relation_drop(Oid relationOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__DROP, true);
	}
#endif
}

void
sepgsql_relation_getattr(Oid relationOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__GETATTR, true);
	}
#endif
}

void
sepgsql_relation_grant(Oid relationOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__SETATTR, true);
	}
#endif
}

void
sepgsql_relation_comment(Oid relationOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__SETATTR, true);
	}
#endif
}

bool
sepgsql_relation_cluster(Oid relationOid, bool abort)
{
#ifdef HAVE_SELINUX
	Assert(get_rel_relkind(relationOid) == RELKIND_RELATION);

	if (sepgsql_is_enabled())
	{
		bool	retval =
			sepgsql_relation_common(relationOid,
									SEPG_DB_TABLE__INDEXON, true);
		return retval;
	}
#endif
	return true;
}

void
sepgsql_relation_truncate(Relation rel)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Assert(RelationGetForm(rel)->relkind == RELKIND_RELATION);

		/* db_table:{delete} */
		sepgsql_relation_common(RelationGetRelid(rel),
								SEPG_DB_TABLE__DELETE, true);
		/* db_tuple:{delete} */
	}
#endif
}

void
sepgsql_relation_lock(Relation rel)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Assert(RelationGetForm(rel)->relkind == RELKIND_RELATION);

		/* db_table:{lock} */
		sepgsql_relation_common(RelationGetRelid(rel),
								SEPG_DB_TABLE__LOCK, true);
	}
#endif
}

void
sepgsql_relation_reindex(Oid relationOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Assert(get_rel_relkind(relationOid) == RELKIND_RELATION);

		/* db_table:{indexon} */
		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__INDEXON, true);
	}
#endif
}

void
sepgsql_view_replace(Oid viewOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Assert(get_rel_relkind(viewOid) == RELKIND_VIEW);
		/* db_view:{setattr} */
		sepgsql_relation_common(viewOid,
								SEPG_DB_VIEW__SETATTR, true);
	}
#endif
}

void
sepgsql_index_create(Oid relationOid, Oid namespaceOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_schema:{add_name} */
		sepgsql_schema_common(namespaceOid,
							  SEPG_DB_SCHEMA__ADD_NAME, true);
		/* db_table:{setattr indexon} */
		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__SETATTR |
								SEPG_DB_TABLE__INDEXON, true);
	}
#endif
}

void
sepgsql_index_reindex(Oid indexOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		HeapTuple	tuple;
		Oid			relationOid;

		tuple = SearchSysCache1(INDEXRELID, indexOid);
		if (!HeapTupleIsValid(tuple))
			elog(ERROR, "cache lookup failed for index %u", indexOid);
		relationOid = ((Form_pg_index) GETSTRUCT(tuple))->indrelid;
		ReleaseSysCache(tuple);

		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__INDEXON, true);
	}
#endif
}

void
sepgsql_sequence_get_value(Oid sequenceOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Assert(get_rel_relkind(sequenceOid) == RELKIND_SEQUENCE);
		sepgsql_relation_common(sequenceOid,
								SEPG_DB_SEQUENCE__GET_VALUE, true);
	}
#endif
}

void
sepgsql_sequence_next_value(Oid sequenceOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Assert(get_rel_relkind(sequenceOid) == RELKIND_SEQUENCE);
		sepgsql_relation_common(sequenceOid,
								SEPG_DB_SEQUENCE__NEXT_VALUE, true);
	}
#endif
}

void
sepgsql_sequence_set_value(Oid sequenceOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Assert(get_rel_relkind(sequenceOid) == RELKIND_SEQUENCE);
		sepgsql_relation_common(sequenceOid,
								SEPG_DB_SEQUENCE__SET_VALUE, true);
	}
#endif
}

void
sepgsql_rule_create(Oid relationOid, const char *ruleName)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__SETATTR, true);
	}
#endif
}

void
sepgsql_rule_drop(Oid relationOid, const char *ruleName, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled() && !cascade)
	{
		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__SETATTR, true);
	}
#endif
}

void
sepgsql_rule_comment(Oid relationOid, const char *ruleName)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__SETATTR, true);
	}
#endif
}

void
sepgsql_trigger_create(Oid relationOid, const char *triggerName,
					   Oid constrrelid, Oid funcOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Assert(get_rel_relkind(relationOid) == RELKIND_RELATION);

		/* db_table:{setattr} */
		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__SETATTR, true);
		/* db_procedure:{install} */
		// sepgsql_procedure_common...
	}
#endif
}

void
sepgsql_trigger_alter(Oid relationOid, const char *triggerName)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Assert(get_rel_relkind(relationOid) == RELKIND_RELATION);
		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__SETATTR, true);
	}
#endif
}

void
sepgsql_trigger_drop(Oid relationOid, const char *triggerName, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled() && !cascade)
	{
		Assert(get_rel_relkind(relationOid) == RELKIND_RELATION);
		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__SETATTR, true);
	}
#endif
}

void
sepgsql_trigger_comment(Oid relationOid, const char *triggerName)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Assert(get_rel_relkind(relationOid) == RELKIND_RELATION);
		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__SETATTR, true);
	}
#endif
}

void
sepgsql_constraint_comment(Oid relationOid, const char *constName)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Assert(get_rel_relkind(relationOid) == RELKIND_RELATION);
		sepgsql_relation_common(relationOid,
								SEPG_DB_TABLE__SETATTR, true);
	}
#endif
}
