/*
 * role.c
 *
 * SELinux hooks related to roles
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "catalog/pg_authid.h"
#include "catalog/pg_seclabel.h"
#include "miscadmin.h"
#include "sepgsql/hooks.h"
#include "sepgsql/sepgsql.h"
#include "utils/lsyscache.h"
#include "utils/syscache.h"

#ifdef HAVE_SELINUX
bool
sepgsql_role_common(Oid roleOid, uint32 required, bool abort)
{
	sepgsql_sid_t	tsid;
	HeapTuple		tuple;
	const char	   *auname;
	bool			retval;

	tuple = SearchSysCache1(AUTHOID, ObjectIdGetDatum(roleOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for role %u", roleOid);

	tsid.relid = AuthIdRelationId;
	tsid.secid = HeapTupleGetSecid(tuple);

	auname = NameStr(((Form_pg_authid) GETSTRUCT(tuple))->rolname);

	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_TUPLE,
								  required,
								  auname,
								  abort);
	ReleaseSysCache(tuple);

	return retval;
}
#endif

Oid
sepgsql_role_create(const char *roleName)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;

		nsid = sepgsql_get_default_tuple_secid(AuthIdRelationId);
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_TUPLE,
							 SEPG_DB_TUPLE__INSERT,
							 roleName,
							 true);
		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_role_alter(Oid roleOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_role_common(roleOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

Oid
sepgsql_role_relabel(Oid roleOid, char *newLabel)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	tsid;

		tsid.relid = AuthIdRelationId;
		tsid.secid = seclabelTransInput(tsid.relid, newLabel);

		/* db_tuple:{update relabelfrom} */
		sepgsql_role_common(roleOid,
							SEPG_DB_TUPLE__UPDATE |
							SEPG_DB_TUPLE__RELABELFROM,
							true);

		/* db_tuple:{relabelto} */
		sepgsql_client_perms(tsid,
							 SEPG_CLASS_DB_TUPLE,
							 SEPG_DB_TUPLE__RELABELTO,
							 GetUserNameFromId(roleOid),
							 true);
		return tsid.secid;
	}
#endif
	ereport(ERROR,
			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
			 errmsg("SE-PostgreSQL is not available")));
	return InvalidOid;
}

void
sepgsql_role_drop(Oid roleOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_role_common(roleOid, SEPG_DB_TUPLE__DELETE, true);
	}
#endif
}

void
sepgsql_role_grant(Oid roleOid, bool is_grant, List *memberIds)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_role_common(roleOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_role_comment(Oid roleOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_role_common(roleOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}
