/*
 * rowlv.c
 *
 * Row-level access control facilities 
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "access/heapam.h"
#include "access/sysattr.h"
#include "catalog/pg_attribute.h"
#include "catalog/pg_class.h"
#include "catalog/pg_database.h"
#include "catalog/pg_language.h"
#include "catalog/pg_largeobject_metadata.h"
#include "catalog/pg_namespace.h"
#include "catalog/pg_proc.h"
#include "catalog/pg_seclabel.h"
#include "catalog/pg_type.h"
#include "nodes/makefuncs.h"
#include "parser/parsetree.h"
#include "sepgsql/hooks.h"
#include "sepgsql/sepgsql.h"
#include "storage/bufmgr.h"
#include "utils/fmgroids.h"
#include "utils/lsyscache.h"
#include "utils/syscache.h"
#include "utils/tqual.h"

static int sepgsql_rowlv_mode = SEPGSQL_ROWLV_FILTER;

int
sepgsql_rowlv_get_mode(void)
{
	return sepgsql_rowlv_mode;
}

int
sepgsql_rowlv_set_mode(int new_mode)
{
	int		old_mode = sepgsql_rowlv_mode;

	Assert(new_mode == SEPGSQL_ROWLV_FILTER ||
		   new_mode == SEPGSQL_ROWLV_ABORT ||
		   new_mode == SEPGSQL_ROWLV_BYPASS);

	sepgsql_rowlv_mode = new_mode;

	return old_mode;
}

uint32
sepgsql_rowlv_permissions(RangeTblEntry *rte)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid			relOid = rte->relid;
		HeapTuple	reltup;
		bool		relhassecids;
		uint32		required = 0;

		if (!OidIsValid(relOid))
			return 0;

		if (get_rel_relkind(relOid) != RELKIND_RELATION)
			return 0;

		reltup = SearchSysCache1(RELOID, ObjectIdGetDatum(relOid));
		if (!HeapTupleIsValid(reltup))
			elog(ERROR, "cache lookup failed for relation %u", relOid);
		relhassecids = ((Form_pg_class) GETSTRUCT(reltup))->relhassecids;
		ReleaseSysCache(reltup);

		if (!relhassecids)
			return 0;

		if (rte->requiredPerms & ACL_SELECT)
			required |= SEPG_DB_TUPLE__SELECT;

		if (rte->requiredPerms & ACL_UPDATE &&
			!bms_is_empty(rte->modifiedCols))
			required |= SEPG_DB_TUPLE__UPDATE;

		if (rte->requiredPerms & ACL_DELETE)
			required |= SEPG_DB_TUPLE__DELETE;

		return required;
	}
#endif
	return 0;
}

#ifdef HAVE_SELINUX
static bool
sepgsql_tuple_perms(Relation rel, HeapTuple tuple, uint32 required, bool abort)
{
	Form_pg_class		relForm;
	Form_pg_attribute	attForm;
	Form_pg_type		typForm;
	sepgsql_sid_t		tsid;
	uint16				tclass;

	/* object class? */
	switch (RelationGetRelid(rel))
	{
		case DatabaseRelationId:
			tclass = SEPG_CLASS_DB_DATABASE;
			break;

		case NamespaceRelationId:
			tclass = SEPG_CLASS_DB_SCHEMA;
			break;

		case RelationRelationId:
			relForm = (Form_pg_class) GETSTRUCT(tuple);
			switch (relForm->relkind)
			{
				case RELKIND_RELATION:
					tclass = SEPG_CLASS_DB_TABLE;
					break;
				case RELKIND_SEQUENCE:
					tclass = SEPG_CLASS_DB_SEQUENCE;
					break;
				case RELKIND_VIEW:
					tclass = SEPG_CLASS_DB_VIEW;
					break;
				case RELKIND_COMPOSITE_TYPE:
					tclass = SEPG_CLASS_DB_TUPLE;
					break;
				default:	/* index, toast */
					tclass = SEPG_CLASS_DB_TABLE;
					break;
			}
			break;

		case AttributeRelationId:
			attForm = (Form_pg_attribute) GETSTRUCT(tuple);
			switch (get_rel_relkind(attForm->attrelid))
			{
				case RELKIND_RELATION:
					tclass = SEPG_CLASS_DB_COLUMN;
					break;
				case RELKIND_SEQUENCE:
					tclass = SEPG_CLASS_DB_SEQUENCE;
					break;
				case RELKIND_VIEW:
					tclass = SEPG_CLASS_DB_VIEW;
					break;
				case RELKIND_COMPOSITE_TYPE:
					tclass = SEPG_CLASS_DB_TUPLE;
					break;
				case RELKIND_INDEX:
				case RELKIND_TOASTVALUE:
					tclass = SEPG_CLASS_DB_TABLE;
					break;
				default:
					/* orphan attribute? */
					return !sepgsql_get_enforce();
			}
			break;

		case ProcedureRelationId:
			tclass = SEPG_CLASS_DB_PROCEDURE;
			break;

		case LanguageRelationId:
			tclass = SEPG_CLASS_DB_LANGUAGE;
			break;

		case LargeObjectMetadataRelationId:
			tclass = SEPG_CLASS_DB_BLOB;
			break;

		case TypeRelationId:
			typForm = (Form_pg_type) GETSTRUCT(tuple);
			if (!OidIsValid(typForm->typrelid))
				tclass = SEPG_CLASS_DB_TUPLE;
			else
			{
				switch (get_rel_relkind(typForm->typrelid))
				{
					case RELKIND_RELATION:
					case RELKIND_INDEX:
					case RELKIND_TOASTVALUE:
						tclass = SEPG_CLASS_DB_TABLE;
						break;
					case RELKIND_SEQUENCE:
						tclass = SEPG_CLASS_DB_SEQUENCE;
						break;
					case RELKIND_VIEW:
						tclass = SEPG_CLASS_DB_VIEW;
						break;
					case RELKIND_COMPOSITE_TYPE:
						tclass = SEPG_CLASS_DB_TUPLE;
						break;
					default:
						/* orphan type? */
						return !sepgsql_get_enforce();
				}
			}
			break;

		default:
			tclass = SEPG_CLASS_DB_TUPLE;
			break;
	}

	/* do permission check */
	tsid.relid = RelationGetRelid(rel);
	tsid.secid = HeapTupleGetSecid(tuple);

	return sepgsql_client_perms(tsid, tclass, required, NULL, abort);
}
#endif

bool
sepgsql_rowlv_quals(Relation rel, uint32 required,
					TupleTableSlot *slot, bool abort)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		HeapTuple	tuple;

		/* no permission to be checked */
		if (!required)
			return true;

		/* mismatch invocation for the current rowlv mode */
		if ((!abort && sepgsql_rowlv_mode != SEPGSQL_ROWLV_FILTER) ||
			(abort && sepgsql_rowlv_mode != SEPGSQL_ROWLV_ABORT))
			return true;

		Assert(OidIsValid(RelationGetRelid(rel)));

		tuple = ExecMaterializeSlot(slot);

		return sepgsql_tuple_perms(rel, tuple, required, abort);
	}
#endif
	return true;
}

bool
sepgsql_rowlv_copyto(Relation rel, HeapTuple tuple)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		return sepgsql_tuple_perms(rel, tuple, SEPG_DB_TUPLE__SELECT, false);
	}
#endif
	return true;
}

void
sepgsql_tuple_insert(Relation rel, HeapTuple tuple)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;
		Oid				relOid = RelationGetRelid(rel);

		if (!RelationGetForm(rel)->relhassecids)
		{
			nsid.relid = RelationRelationId;
			nsid.secid = GetSysCacheSecid1(RELOID, ObjectIdGetDatum(relOid));
		}
		else if (OidIsValid(HeapTupleGetSecid(tuple)))
		{
			nsid.relid = relOid;
			nsid.secid = HeapTupleGetSecid(tuple);
		}
		else
		{
			nsid = sepgsql_get_default_tuple_secid(relOid);
			HeapTupleSetSecid(tuple, nsid.secid);
		}
		/* db_tuple:{insert} */
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_TUPLE,
							 SEPG_DB_TUPLE__INSERT,
							 NULL,
							 true);
	}
#endif
}

void
sepgsql_tuple_update(Relation rel, ItemPointer otid, HeapTuple newtup)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	tsid;
		HeapTupleData	oldtup;
		Buffer			oldbuf;
		Oid				newSecId = HeapTupleGetSecid(newtup);
		Oid				oldSecId;

		/*
		 * heap_update() preserves security id of the original tuple,
		 * if no explicit security label was given, so we don't need
		 * to check anything.
		 * At this point, db_tuple:{update} is already checked.
		 */
		if (!OidIsValid(newSecId))
			return;

		/*
		 * User gave an explicit security label
		 */
		ItemPointerCopy(otid, &oldtup.t_self);
		if (!heap_fetch(rel, SnapshotAny, &oldtup, &oldbuf, false, NULL))
			elog(ERROR, "failed to fetch old version of the tuple");

		tsid.relid = RelationGetRelid(rel);
		oldSecId = HeapTupleGetSecid(&oldtup);

		if (!seclabelCompareSecid(tsid.relid, oldSecId,
								  tsid.relid, newSecId))
		{
			/* db_tuple:{relabelfrom} */
			tsid.secid = oldSecId;
			sepgsql_client_perms(tsid,
								 SEPG_CLASS_DB_TUPLE,
								 SEPG_DB_TUPLE__RELABELFROM,
								 NULL,
								 true);

			/* db_tuple:{relabelto} */
			tsid.secid = newSecId;
			sepgsql_client_perms(tsid,
								 SEPG_CLASS_DB_TUPLE,
								 SEPG_DB_TUPLE__RELABELTO,
								 NULL,
								 true);
		}
		ReleaseBuffer(oldbuf);
	}
#endif
}
