/*
 * schema.c
 *
 * SELinux hooks related to schema
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "catalog/pg_namespace.h"
#include "catalog/pg_seclabel.h"
#include "miscadmin.h"
#include "sepgsql/hooks.h"
#include "sepgsql/sepgsql.h"
#include "utils/lsyscache.h"
#include "utils/syscache.h"

#ifdef HAVE_SELINUX
bool
sepgsql_schema_common(Oid namespaceOid, uint32 required, bool abort)
{
	HeapTuple		tuple;
	sepgsql_sid_t	tsid;
	const char	   *auname;
	bool			retval;

	tuple = SearchSysCache1(NAMESPACEOID,
							ObjectIdGetDatum(namespaceOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for namespace %u", namespaceOid);

	tsid.relid = NamespaceRelationId;
	tsid.secid = HeapTupleGetSecid(tuple);

	auname = NameStr(((Form_pg_namespace) GETSTRUCT(tuple))->nspname);
	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_SCHEMA,
								  required,
								  auname,
								  abort);
	ReleaseSysCache(tuple);

	return retval;
}
#endif

Oid
sepgsql_schema_create(const char *nspName, bool is_temp)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;

		/* compute default security context */
		nsid = sepgsql_get_default_schema_secid(MyDatabaseId);

		/* db_schema:{create} */
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_SCHEMA,
							 SEPG_DB_SCHEMA__CREATE,
							 nspName, true);
		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_schema_alter(Oid namespaceOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_schema_common(namespaceOid,
							  SEPG_DB_SCHEMA__SETATTR,
							  true);
	}
#endif
}

Oid
sepgsql_schema_relabel(Oid namespaceOid, char *new_label)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;
		char		   *auname;

		nsid.relid = NamespaceRelationId;
		nsid.secid = seclabelTransInput(nsid.relid, new_label);

		auname = get_namespace_name(namespaceOid);

		/* db_schema:{setattr relabelfrom} */
		sepgsql_schema_common(namespaceOid,
							  SEPG_DB_SCHEMA__SETATTR |
							  SEPG_DB_SCHEMA__RELABELFROM,
							  true);

		/* db_schema:{relabelto} */
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_SCHEMA,
							 SEPG_DB_SCHEMA__RELABELTO,
							 auname,
							 true);
		pfree(auname);

		return nsid.secid;
	}
#endif
	ereport(ERROR,
			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
			 errmsg("SE-PostgreSQL is not available")));
	return InvalidOid;
}

void
sepgsql_schema_drop(Oid namespaceOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_schema_common(namespaceOid,
							  SEPG_DB_SCHEMA__DROP,
							  true);
	}
#endif
}

void
sepgsql_schema_grant(Oid namespaceOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_schema_common(namespaceOid,
							  SEPG_DB_SCHEMA__SETATTR,
							  true);
	}
#endif
}

bool
sepgsql_schema_search(Oid namespaceOid, bool abort)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		bool	retval =
			sepgsql_schema_common(namespaceOid,
								  SEPG_DB_SCHEMA__SEARCH,
								  abort);
		return retval;
	}
#endif
	return true;
}

void
sepgsql_schema_comment(Oid namespaceOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_schema_common(namespaceOid,
							  SEPG_DB_SCHEMA__SETATTR,
							  true);
	}
#endif
}
