/*
 * src/backend/security/sepgsql/selinux.c
 *   Routines to communicate with SELinux.
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "catalog/pg_seclabel.h"
#include "libpq/libpq.h"
#include "miscadmin.h"
#include "sepgsql/sepgsql.h"
#include "utils/guc.h"
#include "utils/memutils.h"

#include <selinux/selinux.h>

/*
 * selinux_catalog
 *
 * This static translation lookup table enables to associate a certain
 * object class/permission name with its internal code, such as
 * SEPG_CLASS_DB_SCHEMA.
 *
 * SELinux requires applications to represent object class and a set of
 * permissions in code, instead of its name, when we ask SELinux's decision.
 *
 * See the definition of security_compute_av(3) API in libselinux.
 * We need to gives a code of object class, and interpret what permissions
 * are allowed on the object class from av_decision structure.
 * Actual values of the code depend on the security policy. In other words,
 * we cannot know what number is assigned on a certain object class and
 * permissions.
 * The string_to_security_class(3) and string_to_av_perm(3) APIs takes
 * arguments with the name of object class/permission, and returns the
 * code for the given object class/permissions.
 * For example, we can know what code is assigned on the "db_table" class
 * using these functions as follows:
 *
 *   uint16 tclass_ex = string_to_security_class("db_table");
 *
 * On the other hand, we use an alternative code internally to simplify
 * the implementation, such as SEPG_CLASS_* for object class.
 * The following selinux_catalog is used to translate the 'internal'
 * code and the 'external' code.
 *
 * It allows to lookup name of the object class or permission corresponding
 * to a certain 'internal' code. Then, we can give the name to SELinux's
 * API to obtain 'external' code which can be used to ask in-kernel SELinux.
 */
static struct
{
	const char	   *class_name;
	uint16			class_code;
	struct
	{
		const char *perm_name;
		uint32		perm_code;
	} perms[32];
} selinux_catalog[] = {
	{
		"process",				SEPG_CLASS_PROCESS,
		{
			{ "translation",	SEPG_PROCESS__TRANSITION },
			{ NULL, 0UL }
		}
	},
	{
		"file",					SEPG_CLASS_FILE,
		{
			{ "read",			SEPG_FILE__READ },
			{ "write",			SEPG_FILE__WRITE },
			{ "create",			SEPG_FILE__CREATE },
			{ "getattr",		SEPG_FILE__GETATTR },
			{ "unlink",			SEPG_FILE__UNLINK },
			{ "rename",			SEPG_FILE__RENAME },
			{ "append",			SEPG_FILE__APPEND },
			{ NULL, 0UL }
		}
	},
	{
		"dir",					SEPG_CLASS_DIR,
		{
			{ "read",			SEPG_DIR__READ },
			{ "write",			SEPG_DIR__WRITE },
			{ "create",			SEPG_DIR__CREATE },
			{ "getattr",		SEPG_DIR__GETATTR },
			{ "unlink",			SEPG_DIR__UNLINK },
			{ "rename",			SEPG_DIR__RENAME },
			{ "search",			SEPG_DIR__SEARCH },
			{ "add_name",		SEPG_DIR__ADD_NAME },
			{ "remove_name",	SEPG_DIR__REMOVE_NAME },
			{ "rmdir",			SEPG_DIR__RMDIR },
			{ "reparent",		SEPG_DIR__REPARENT },
			{ NULL, 0UL }
		}
	},
	{
		"lnk_file",				SEPG_CLASS_LNK_FILE,
		{
			{ "read",			SEPG_LNK_FILE__READ },
			{ "write",			SEPG_LNK_FILE__WRITE },
			{ "create",			SEPG_LNK_FILE__CREATE },
			{ "getattr",		SEPG_LNK_FILE__GETATTR },
			{ "unlink",			SEPG_LNK_FILE__UNLINK },
			{ "rename",			SEPG_LNK_FILE__RENAME },
			{ NULL, 0UL }
		}
	},
	{
		"chr_file",				SEPG_CLASS_CHR_FILE,
		{
			{ "read",			SEPG_CHR_FILE__READ },
			{ "write",			SEPG_CHR_FILE__WRITE },
			{ "create",			SEPG_CHR_FILE__CREATE },
			{ "getattr",		SEPG_CHR_FILE__GETATTR },
			{ "unlink",			SEPG_CHR_FILE__UNLINK },
			{ "rename",			SEPG_CHR_FILE__RENAME },
			{ NULL, 0UL }
		}
	},
	{
		"blk_file",				SEPG_CLASS_BLK_FILE,
		{
			{ "read",			SEPG_BLK_FILE__READ },
			{ "write",			SEPG_BLK_FILE__WRITE },
			{ "create",			SEPG_BLK_FILE__CREATE },
			{ "getattr",		SEPG_BLK_FILE__GETATTR },
			{ "unlink",			SEPG_BLK_FILE__UNLINK },
			{ "rename",			SEPG_BLK_FILE__RENAME },
			{ NULL, 0UL }
		}
	},
	{
		"sock_file",			SEPG_CLASS_SOCK_FILE,
		{
			{ "read",			SEPG_SOCK_FILE__READ },
			{ "write",			SEPG_SOCK_FILE__WRITE },
			{ "create",			SEPG_SOCK_FILE__CREATE },
			{ "getattr",		SEPG_SOCK_FILE__GETATTR },
			{ "unlink",			SEPG_SOCK_FILE__UNLINK },
			{ "rename",			SEPG_SOCK_FILE__RENAME },
			{ NULL, 0UL }
		}
	},
	{
		"fifo_file",			SEPG_CLASS_FIFO_FILE,
		{
			{ "read",			SEPG_FIFO_FILE__READ },
			{ "write",			SEPG_FIFO_FILE__WRITE },
			{ "create",			SEPG_FIFO_FILE__CREATE },
			{ "getattr",		SEPG_FIFO_FILE__GETATTR },
			{ "unlink",			SEPG_FIFO_FILE__UNLINK },
			{ "rename",			SEPG_FIFO_FILE__RENAME },
			{ NULL, 0UL }
		}
	},
	{
		"db_database",			SEPG_CLASS_DB_DATABASE,
		{
			{ "create",			SEPG_DB_DATABASE__CREATE },
			{ "drop",			SEPG_DB_DATABASE__DROP },
			{ "getattr",		SEPG_DB_DATABASE__GETATTR },
			{ "setattr",		SEPG_DB_DATABASE__SETATTR },
			{ "relabelfrom",	SEPG_DB_DATABASE__RELABELFROM },
			{ "relabelto",		SEPG_DB_DATABASE__RELABELTO },
			{ "access",			SEPG_DB_DATABASE__ACCESS },
			{ "load_module",	SEPG_DB_DATABASE__LOAD_MODULE },
			{ NULL, 0UL },
		}
	},
	{
		"db_schema",			SEPG_CLASS_DB_SCHEMA,
		{
			{ "create",			SEPG_DB_SCHEMA__CREATE },
			{ "drop",			SEPG_DB_SCHEMA__DROP },
			{ "getattr",		SEPG_DB_SCHEMA__GETATTR },
			{ "setattr",		SEPG_DB_SCHEMA__SETATTR },
			{ "relabelfrom",	SEPG_DB_SCHEMA__RELABELFROM },
			{ "relabelto",		SEPG_DB_SCHEMA__RELABELTO },
			{ "search",			SEPG_DB_SCHEMA__SEARCH },
			{ "add_name",		SEPG_DB_SCHEMA__ADD_NAME },
			{ "remove_name",	SEPG_DB_SCHEMA__REMOVE_NAME },
			{ NULL, 0UL },
		}
	},
	{
		"db_table",             SEPG_CLASS_DB_TABLE,
		{
			{ "create",         SEPG_DB_TABLE__CREATE },
			{ "drop",           SEPG_DB_TABLE__DROP },
			{ "getattr",        SEPG_DB_TABLE__GETATTR },
			{ "setattr",        SEPG_DB_TABLE__SETATTR },
			{ "relabelfrom",    SEPG_DB_TABLE__RELABELFROM },
			{ "relabelto",      SEPG_DB_TABLE__RELABELTO },
			{ "select",         SEPG_DB_TABLE__SELECT },
			{ "update",         SEPG_DB_TABLE__UPDATE },
			{ "insert",         SEPG_DB_TABLE__INSERT },
			{ "delete",         SEPG_DB_TABLE__DELETE },
			{ "lock",           SEPG_DB_TABLE__LOCK },
			{ "indexon",		SEPG_DB_TABLE__INDEXON },
			{ NULL, 0UL },
		}
	},
	{
		"db_view",				SEPG_CLASS_DB_VIEW,
		{
			{ "create",			SEPG_DB_VIEW__CREATE },
			{ "drop",			SEPG_DB_VIEW__DROP },
			{ "getattr",		SEPG_DB_VIEW__GETATTR },
			{ "setattr",		SEPG_DB_VIEW__SETATTR },
			{ "relabelfrom",	SEPG_DB_VIEW__RELABELFROM },
			{ "relabelto",		SEPG_DB_VIEW__RELABELTO },
			{ "expand",			SEPG_DB_VIEW__EXPAND },
			{ NULL, 0UL },
		}
	},
	{
		"db_sequence",			SEPG_CLASS_DB_SEQUENCE,
		{
			{ "create",			SEPG_DB_SEQUENCE__CREATE },
			{ "drop",			SEPG_DB_SEQUENCE__DROP },
			{ "getattr",		SEPG_DB_SEQUENCE__GETATTR },
			{ "setattr",		SEPG_DB_SEQUENCE__SETATTR },
			{ "relabelfrom",	SEPG_DB_SEQUENCE__RELABELFROM },
			{ "relabelto",		SEPG_DB_SEQUENCE__RELABELTO },
			{ "get_value",		SEPG_DB_SEQUENCE__GET_VALUE },
			{ "next_value",		SEPG_DB_SEQUENCE__NEXT_VALUE },
			{ "set_value",		SEPG_DB_SEQUENCE__SET_VALUE },
			{ NULL, 0UL },
		}
	},
	{
		"db_procedure",			SEPG_CLASS_DB_PROCEDURE,
		{
			{ "create",			SEPG_DB_PROCEDURE__CREATE },
			{ "drop",			SEPG_DB_PROCEDURE__DROP },
			{ "getattr",		SEPG_DB_PROCEDURE__GETATTR },
			{ "setattr",		SEPG_DB_PROCEDURE__SETATTR },
			{ "relabelfrom",	SEPG_DB_PROCEDURE__RELABELFROM },
			{ "relabelto",		SEPG_DB_PROCEDURE__RELABELTO },
			{ "execute",		SEPG_DB_PROCEDURE__EXECUTE },
			{ "entrypoint",		SEPG_DB_PROCEDURE__ENTRYPOINT },
			{ "install",		SEPG_DB_PROCEDURE__INSTALL },
			{ NULL, 0UL },
		}
	},
	{
		"db_column",			SEPG_CLASS_DB_COLUMN,
		{
			{ "create",			SEPG_DB_COLUMN__CREATE },
			{ "drop",			SEPG_DB_COLUMN__DROP },
			{ "getattr",		SEPG_DB_COLUMN__GETATTR },
			{ "setattr",		SEPG_DB_COLUMN__SETATTR },
			{ "relabelfrom",	SEPG_DB_COLUMN__RELABELFROM },
			{ "relabelto",		SEPG_DB_COLUMN__RELABELTO },
			{ "select",			SEPG_DB_COLUMN__SELECT },
			{ "update",			SEPG_DB_COLUMN__UPDATE },
			{ "insert",			SEPG_DB_COLUMN__INSERT },
			{ NULL, 0UL },
		}
	},
	{
		"db_tuple",				SEPG_CLASS_DB_TUPLE,
		{
			{ "relabelfrom",	SEPG_DB_TUPLE__RELABELFROM },
			{ "relabelto",		SEPG_DB_TUPLE__RELABELTO },
			{ "select",			SEPG_DB_TUPLE__SELECT },
			{ "update",			SEPG_DB_TUPLE__UPDATE },
			{ "insert",			SEPG_DB_TUPLE__INSERT },
			{ "delete",			SEPG_DB_TUPLE__DELETE },
			{ NULL, 0UL },
		}
	},
	{
		"db_blob",				SEPG_CLASS_DB_BLOB,
		{
			{ "create",			SEPG_DB_BLOB__CREATE },
			{ "drop",			SEPG_DB_BLOB__DROP },
			{ "getattr",		SEPG_DB_BLOB__GETATTR },
			{ "setattr",		SEPG_DB_BLOB__SETATTR },
			{ "relabelfrom",	SEPG_DB_BLOB__RELABELFROM },
			{ "relabelto",		SEPG_DB_BLOB__RELABELTO },
			{ "read",			SEPG_DB_BLOB__READ },
			{ "write",			SEPG_DB_BLOB__WRITE },
			{ "import",			SEPG_DB_BLOB__IMPORT },
			{ "export",			SEPG_DB_BLOB__EXPORT },
			{ NULL, 0UL },
		}
	},
	{
		"db_language",			SEPG_CLASS_DB_LANGUAGE,
		{
			{ "create",			SEPG_DB_LANGUAGE__CREATE },
			{ "drop",			SEPG_DB_LANGUAGE__DROP },
			{ "getattr",		SEPG_DB_LANGUAGE__GETATTR },
			{ "setattr",		SEPG_DB_LANGUAGE__SETATTR },
			{ "relabelfrom",	SEPG_DB_LANGUAGE__RELABELFROM },
			{ "relabelto",		SEPG_DB_LANGUAGE__RELABELTO },
			{ "implement",		SEPG_DB_LANGUAGE__IMPLEMENTE },
			{ "execute",		SEPG_DB_LANGUAGE__EXECUTE },
			{ NULL, 0UL },
		}
	},
};

/*
 * GUC option: sepostgresql = [default|enforcing|permissive|disabled]
 *
 * SEPGSQL_MODE_DEFAULT		: It follows system setting
 * SEPGSQL_MODE_ENFORCING	: Use enforcing mode always
 * SEPGSQL_MODE_PERMISSIVE	: Use permissive mode always
 * SEPGSQL_MODE_INTERNAL	: Internally used mode. Same as permissive mode
 *							  except for silence in audit logs
 * SEPGSQL_MODE_DISABLED	: It always disables SE-PgSQL configuration
 */
int sepostgresql_mode;

/*
 * sepgsql_is_enabled
 *
 * If it returns true, SE-PgSQL is enabled. Otherwise, it is disabled.
 */
bool
sepgsql_is_enabled(void)
{
	static int	enabled = -1;

	/*
	 * If sepostgresql = disabled, it always returns FALSE
	 * independently from the system status.
	 */
	if (sepostgresql_mode == SEPGSQL_MODE_DISABLED)
		return false;

	/*
	 * SE-PgSQL needs SELinux is enabled on the operating system.
	 * If it is disabled, SE-PgSQL has to be also disabled, even if
	 * 'enforcing' or 'permissive' are specified.
	 */
	if (enabled < 0)
		enabled = is_selinux_enabled();

	return enabled > 0 ? true : false;
}

/*
 * sepgsql_get_enforce
 *
 * It returns true, if SE-PgSQL performs in enforcing mode.
 *
 * In enforcing mode, SE-PgSQL performs as expected. It checks permissions
 * on the required action, and it prevents them if violated.
 * In permissive mode, SE-PgSQL also checks permissions, but it does not
 * prevent anything, even if violated. It generates audit logs for access
 * violations, so we can use this mode to debug security policy itself.
 */
bool
sepgsql_get_enforce(void)
{
	if (sepostgresql_mode == SEPGSQL_MODE_DEFAULT)
	{
		if (security_getenforce() == 1)
			return true;
	}
	else if (sepostgresql_mode == SEPGSQL_MODE_ENFORCING)
		return true;

	return false;
}

/*
 * sepgsql_show_mode
 *
 * It returns the current performing mode ('selinux_support')
 * in human readable form.
 */
const char *
sepgsql_show_mode(void)
{
	if (!sepgsql_is_enabled())
		return "disabled";

	if (!sepgsql_get_enforce())
		return "permissive";

	return "enforcing";
}

/*
 * GUC parameter to turn on/off debuging audit generation
 */
bool	sepgsql_debug_audit;

/*
 * sepgsql_audit_log
 *
 * It generates a security audit record. In the default, it writes out
 * audit records into standard PG's logfile. It also allows to set up
 * external audit log receiver, such as auditd in Linux, using the
 * sepgsql_audit_hook.
 *
 * SELinux can control what should be audited and should not using
 * "auditdeny" and "auditallow" rules in the security policy. In the
 * default, all the access violations are audited, and all the access
 * allowed are not audited. But we can set up the security policy, so
 * we can have exceptions. So, it is necessary to follow the suggestion
 * come from the security policy. (av_decision.auditallow and auditdeny)
 *
 * Security audit is an important feature, because it enables us to check
 * what was happen if we have a security incident. In fact, ISO/IEC15408
 * defines several security functionalities for audit features.
 */
void
sepgsql_audit_log(bool denied, char *scontext, char *tcontext,
				  uint16 tclass, uint32 audited, const char *audit_name)
{
	StringInfoData	buf;
	const char	   *tclass_name;
	const char	   *perm_name;
	int				level = LOG;
	int				i;

	/*
	 * translation of security contexts to human readable format,
	 * if sepgsql_mcstrans is turned on.
	 */
	scontext = sepgsql_mcstrans_out(scontext);
	tcontext = sepgsql_mcstrans_out(tcontext);

	/* lookup name of the object class */
	tclass_name = selinux_catalog[tclass].class_name;

	/* lookup name of the permissions */
	initStringInfo(&buf);
	appendStringInfo(&buf, "{");

	for (i=0; selinux_catalog[tclass].perms[i].perm_name; i++)
	{
		if (audited & (1UL << i))
		{
			perm_name = selinux_catalog[tclass].perms[i].perm_name;
			appendStringInfo(&buf, " %s", perm_name);
		}
	}
	appendStringInfo(&buf, " }");

	/*
	 * Call external audit module, if loaded
	 */
	appendStringInfo(&buf, " scontext=%s tcontext=%s tclass=%s",
					 scontext, tcontext, tclass_name);
	if (audit_name)
		appendStringInfo(&buf, " name=%s", audit_name);

	if (sepgsql_debug_audit)
		level = client_min_messages;

	ereport(level,
			(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
			 errmsg("SELinux: %s %s",
					(denied ? "denied" : "allowed"), buf.data)));
}

/*
 * sepgsql_compute_avd
 *
 * It actually asks SELinux what permissions are allowed on a pair of
 * the security contexts and object class. It also returns what permissions
 * should be audited on access violation or allowed.
 * In most cases, subject's security context (scontext) is a client, and
 * target security context (tcontext) is a database object.
 *
 * The access control decision shall be set on the given av_decision.
 * The av_decision.allowed has a bitmask of SEPG_<class>__<perms>
 * to suggest a set of allowed actions in this object class.
 */
void
sepgsql_compute_avd(char *scontext, char *tcontext,
					uint16 tclass, struct av_decision *avd)
{
	const char		   *tclass_name;
	security_class_t	tclass_ex;
	struct av_decision	avd_ex;
	int					i, deny_unknown = security_deny_unknown();

	/* Get external code of the object class*/
	Assert(tclass < SEPG_CLASS_MAX);

	tclass_name = selinux_catalog[tclass].class_name;
	tclass_ex = string_to_security_class(tclass_name);

	if (tclass_ex == 0)
	{
		/*
		 * If the current security policy does not support permissions
		 * corresponding to database objects, we fill up them with dummy
		 * data.
		 * If security_deny_unknown() returns positive value, undefined
		 * permissions should be denied. Otherwise, allowed
		 */
		avd->allowed = (deny_unknown > 0 ? 0 : ~0U);
		avd->auditallow = 0U;
		avd->auditdeny =  ~0U;
		avd->flags = 0;

		return;
	}

	/*
	 * Ask SELinux what is allowed set of permissions on a pair of the
	 * security contexts and the given object class.
	 */
	if (security_compute_av_flags_raw(scontext, tcontext,
									  tclass_ex, 0, &avd_ex) < 0)
		ereport(ERROR,
				(errcode(ERRCODE_INTERNAL_ERROR),
				 errmsg("SELinux could not compute av_decision: "
						"scontext=%s tcontext=%s tclass=%s",
						scontext, tcontext, tclass_name)));

	/*
	 * SELinux returns its access control decision as a set of permissions
	 * represented in external code which depends on run-time environment.
	 * So, we need to translate it to the internal representation before
	 * returning results for the caller.
	 */
	memset(avd, 0, sizeof(struct av_decision));

	for (i=0; selinux_catalog[tclass].perms[i].perm_name; i++)
	{
		access_vector_t		perm_code_ex;
		const char *perm_name = selinux_catalog[tclass].perms[i].perm_name;
		uint32		perm_code = selinux_catalog[tclass].perms[i].perm_code;

		perm_code_ex = string_to_av_perm(tclass_ex, perm_name);
		if (perm_code_ex == 0)
		{
			/* fill up undefined permissions */
			if (!deny_unknown)
				avd->allowed |= perm_code;
			avd->auditdeny |= perm_code;

			continue;
		}

		if (avd_ex.allowed & perm_code_ex)
			avd->allowed |= perm_code;
		if (avd_ex.auditallow & perm_code_ex)
			avd->auditallow |= perm_code;
		if (avd_ex.auditdeny & perm_code_ex)
			avd->auditdeny |= perm_code;
	}

	return;
}

/*
 * sepgsql_compute_perms
 *
 * It makes access control decision communicating with SELinux.
 * If SELinux does not allow required permissions on a pair of the security
 * contexts, it raises an error or returns false.
 *
 * scontext : The security context of subject. In most cases, it is client.
 * tcontext : The security context of target database object.
 * tclass   : One of the object class code (SEPG_CLASS_*) declared in the
 *            header file.
 * required : A bitmap of the required permissions (SEPG_<class>__<perm>)
 *            declared in the header file.
 * audit_name : A human readable name of the database object for auditing.
 * abort    : True, if caller want to raise an error on access violation.
 */
bool
sepgsql_compute_perms(char *scontext, char *tcontext,
                      uint16 tclass, uint32 required,
                      const char *audit_name, bool abort)
{
	struct av_decision	avd;
	uint32		denied;
	uint32		audited;

	sepgsql_compute_avd(scontext, tcontext, tclass, &avd);

	/*
	 * It logs a security audit record for the given request, if necessary.
	 * When SE-PgSQL performs 'internal' mode, it needs to keep silent.
	 */
	denied = required & ~avd.allowed;
	if (sepgsql_debug_audit && tclass != SEPG_CLASS_DB_TUPLE)
		audited = (denied ? (denied & ~0) : (required & ~0));
	else
		audited = (denied ? (denied & avd.auditdeny)
						  : (required & avd.auditallow));

	if (audited && sepostgresql_mode != SEPGSQL_MODE_INTERNAL)
	{
		sepgsql_audit_log(!!denied, scontext, tcontext,
						  tclass, audited, audit_name);
	}

	/*
	 * If here is no policy violations, or SE-PgSQL performs in permissive
	 * mode, or the client process peforms in permissive domain, it returns
	 * normally with 'true'.
	 */
	if (!denied ||
		!sepgsql_get_enforce() ||
		(avd.flags & SELINUX_AVD_FLAGS_PERMISSIVE) != 0)
		return true;

	/*
	 * Otherwise, it raises an error or returns 'false', depending on the
	 * caller's indication by 'abort'.
	 */
	if (abort)
		ereport(ERROR,
				(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
				 errmsg("SELinux: security policy violation")));

	return false;
}

/*
 * sepgsql_compute_create
 *
 * It returns a default security context to be assigned on a new database
 * object. SELinux compute it based on a combination of client, upper object
 * which owns the new object and object class.
 *
 * For example, when a client (staff_u:staff_r:staff_t:s0) tries to create
 * a new table within a schema (system_u:object_r:sepgsql_schema_t:s0),
 * SELinux looks-up its security policy. If it has a special rule on the
 * combination of these security contexts and object class (db_table),
 * it returns the security context suggested by the special rule.
 * Otherwise, it returns the security context of schema, as is.
 *
 * We expect the caller already applies sanity/validation checks on the
 * given security context.
 *
 * scontext : The security context of subject. In most cases, it is client.
 * tcontext : The security context of the parent database object..
 * tclass   : One of the object class code (SEPG_CLASS_*) declared in the
 *            header file.
 */
char *
sepgsql_compute_create(char *scontext, char *tcontext, uint16 tclass)
{
	security_context_t	ncontext;
	security_class_t	tclass_ex;
	const char		   *tclass_name;
	char			   *result;

	/* Get external code of the object class*/
	Assert(tclass < SEPG_CLASS_MAX);

	tclass_name = selinux_catalog[tclass].class_name;
	tclass_ex = string_to_security_class(tclass_name);

	/*
	 * Ask SELinux what is the default context for the given object class
	 * on a pair of security contexts
	 */
	if (security_compute_create_raw(scontext, tcontext,
									tclass_ex, &ncontext))
		ereport(ERROR,
				(errcode(ERRCODE_INTERNAL_ERROR),
				 errmsg("SELinux could not compute a new context: "
						"scontext=%s tcontext=%s tclass=%s",
						scontext, tcontext, tclass_name)));

	/*
	 * libselinux returns malloc()'ed string, so we need to copy it
	 * on the palloc()'ed region.
	 */
	PG_TRY();
	{
		result = pstrdup(ncontext);
	}
	PG_CATCH();
	{
		freecon(ncontext);
		PG_RE_THROW();
	}
	PG_END_TRY();
	freecon(ncontext);

	return result;
}
