/*
 * tablespace.c
 *
 * SELinux hooks related to tablespaces
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "catalog/pg_tablespace.h"
#include "catalog/pg_seclabel.h"
#include "commands/tablespace.h"
#include "sepgsql/hooks.h"
#include "sepgsql/sepgsql.h"
#include "utils/syscache.h"
#include "utils/lsyscache.h"

#ifdef HAVE_SELINUX
bool
sepgsql_tablespace_common(Oid tablespaceOid, uint32 required, bool abort)
{
	HeapTuple		tuple;
	sepgsql_sid_t	tsid;
	bool			retval;
	const char	   *auname;

	tuple = SearchSysCache1(TABLESPACEOID, ObjectIdGetDatum(tablespaceOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for tablespace %u", tablespaceOid);

	tsid.relid = TableSpaceRelationId;
	tsid.secid = HeapTupleGetSecid(tuple);

	auname = NameStr(((Form_pg_tablespace) GETSTRUCT(tuple))->spcname);
	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_TUPLE,
								  required,
								  auname,
								  abort);
	ReleaseSysCache(tuple);

	return retval;
}

#endif

Oid
sepgsql_tablespace_create(const char *tablespaceName)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;

		nsid = sepgsql_get_default_tuple_secid(TableSpaceRelationId);
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_TUPLE,
							 SEPG_DB_TUPLE__INSERT,
							 tablespaceName,
							 true);
		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_tablespace_alter(Oid tablespaceOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_tablespace_common(tablespaceOid,
								  SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

Oid
sepgsql_tablespace_relabel(Oid tablespaceOid, char *newLabel)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	tsid;

		tsid.relid = TableSpaceRelationId;
		tsid.secid = seclabelTransInput(tsid.relid, newLabel);

		/* db_tuple:{update relabelfrom} */
		sepgsql_tablespace_common(tablespaceOid,
								  SEPG_DB_TUPLE__UPDATE |
								  SEPG_DB_TUPLE__RELABELFROM, true);

		/* db_procedure:{relabelto} */
		sepgsql_client_perms(tsid,
							 SEPG_CLASS_DB_TUPLE,
							 SEPG_DB_TUPLE__RELABELTO,
							 get_tablespace_name(tablespaceOid),
							 true);
		return tsid.secid;
	}
#endif
	ereport(ERROR,
			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
			 errmsg("SE-PostgreSQL is not available")));
    return InvalidOid;
}

void
sepgsql_tablespace_drop(Oid tablespaceOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_tablespace_common(tablespaceOid,
								  SEPG_DB_TUPLE__DELETE, true);
	}
#endif
}

void
sepgsql_tablespace_grant(Oid tablespaceOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_tablespace_common(tablespaceOid,
								  SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_tablespace_getattr(Oid tablespaceOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_tablespace_common(tablespaceOid,
								  SEPG_DB_TUPLE__SELECT, true);
	}
#endif
}

void
sepgsql_tablespace_comment(Oid tablespaceOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_tablespace_common(tablespaceOid,
								  SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}
