/*
 * tsearch.c
 *
 * SELinux hooks related to text searches
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "catalog/pg_ts_config.h"
#include "catalog/pg_ts_dict.h"
#include "catalog/pg_ts_parser.h"
#include "catalog/pg_ts_template.h"
#include "sepgsql/hooks.h"
#include "sepgsql/sepgsql.h"
#include "utils/syscache.h"

#ifdef HAVE_SELINUX
bool
sepgsql_ts_config_common(Oid confOid, uint32 required, bool abort)
{
	sepgsql_sid_t	tsid;
	HeapTuple		tuple;
	const char	   *auname;
	bool			retval;

	tuple = SearchSysCache1(TSCONFIGOID, ObjectIdGetDatum(confOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for text search configuration %u", confOid);

	tsid.relid = TSConfigRelationId;
	tsid.secid = HeapTupleGetSecid(tuple);

	auname = NameStr(((Form_pg_ts_config) GETSTRUCT(tuple))->cfgname);

	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_TUPLE,
								  required,
								  auname,
								  abort);
	ReleaseSysCache(tuple);

	return retval;
}

static Oid
get_ts_config_namespace(Oid confOid)
{
	HeapTuple	tuple;
	Oid			namespaceId = InvalidOid;

	tuple = SearchSysCache1(TSCONFIGOID, ObjectIdGetDatum(confOid));
	if (HeapTupleIsValid(tuple))
	{
		namespaceId = ((Form_pg_ts_config) GETSTRUCT(tuple))->cfgnamespace;

		ReleaseSysCache(tuple);
	}
	return namespaceId;
}

bool
sepgsql_ts_dict_common(Oid dictOid, uint32 required, bool abort)
{
	sepgsql_sid_t	tsid;
	HeapTuple		tuple;
	const char	   *auname;
	bool			retval;

	tuple = SearchSysCache1(TSDICTOID, ObjectIdGetDatum(dictOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for text search dictionary %u", dictOid);

	tsid.relid = TSDictionaryRelationId;
	tsid.secid = HeapTupleGetSecid(tuple);

	auname = NameStr(((Form_pg_ts_dict) GETSTRUCT(tuple))->dictname);

	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_TUPLE,
								  required,
								  auname,
								  abort);
	ReleaseSysCache(tuple);

	return retval;
}

static Oid
get_ts_dict_namespace(Oid dictOid)
{
	HeapTuple	tuple;
	Oid			namespaceId = InvalidOid;

	tuple = SearchSysCache1(TSDICTOID, ObjectIdGetDatum(dictOid));
	if (HeapTupleIsValid(tuple))
	{
		namespaceId = ((Form_pg_ts_dict) GETSTRUCT(tuple))->dictnamespace;

		ReleaseSysCache(tuple);
	}
	return namespaceId;
}

bool
sepgsql_ts_parser_common(Oid parseOid, uint32 required, bool abort)
{
	sepgsql_sid_t	tsid;
	HeapTuple		tuple;
	const char	   *auname;
	bool			retval;

	tuple = SearchSysCache1(TSPARSEROID, ObjectIdGetDatum(parseOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for text search parser %u", parseOid);

	tsid.relid = TSParserRelationId;
	tsid.secid = HeapTupleGetSecid(tuple);

	auname = NameStr(((Form_pg_ts_parser) GETSTRUCT(tuple))->prsname);

	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_TUPLE,
								  required,
								  auname,
								  abort);
	ReleaseSysCache(tuple);

	return retval;
}

static Oid
get_ts_parser_namespace(Oid parseOid)
{
	HeapTuple	tuple;
	Oid			namespaceId = InvalidOid;

	tuple = SearchSysCache1(TSPARSEROID, ObjectIdGetDatum(parseOid));
	if (HeapTupleIsValid(tuple))
	{
		namespaceId = ((Form_pg_ts_parser) GETSTRUCT(tuple))->prsnamespace;

		ReleaseSysCache(tuple);
	}
	return namespaceId;
}

bool
sepgsql_ts_template_common(Oid templateOid, uint32 required, bool abort)
{
	sepgsql_sid_t	tsid;
	HeapTuple		tuple;
	const char	   *auname;
	bool			retval;

	tuple = SearchSysCache1(TSTEMPLATEOID, ObjectIdGetDatum(templateOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for text search template %u", templateOid);

	tsid.relid = TSDictionaryRelationId;
	tsid.secid = HeapTupleGetSecid(tuple);

	auname = NameStr(((Form_pg_ts_dict) GETSTRUCT(tuple))->dictname);

	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_TUPLE,
								  required,
								  auname,
								  abort);
	ReleaseSysCache(tuple);

	return retval;
}

static Oid
get_ts_template_namespace(Oid templateOid)
{
	HeapTuple	tuple;
	Oid			namespaceId = InvalidOid;

	tuple = SearchSysCache1(TSTEMPLATEOID, ObjectIdGetDatum(templateOid));
	if (HeapTupleIsValid(tuple))
	{
		namespaceId = ((Form_pg_ts_template) GETSTRUCT(tuple))->tmplnamespace;

		ReleaseSysCache(tuple);
	}
	return namespaceId;
}

#endif

Oid
sepgsql_ts_config_create(const char *confName, Oid namespaceId)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid
			= sepgsql_get_default_tuple_secid(TSConfigRelationId);

		/* db_schema:{add_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__ADD_NAME, true);

		/* db_tuple:{insert} */
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_TUPLE,
							 SEPG_DB_TUPLE__INSERT,
							 confName,
							 true);
		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_ts_config_alter(Oid confOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_ts_config_common(confOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_ts_config_alter_rename(Oid confOid, const char *newName)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		namespaceId = get_ts_config_namespace(confOid);

		/* db_schema:{add_name remove_name} */
		sepgsql_schema_common(namespaceId,
							  SEPG_DB_SCHEMA__ADD_NAME |
							  SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_tuple:{update} */
		sepgsql_ts_config_common(confOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_ts_config_drop(Oid confOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		namespaceId = get_ts_config_namespace(confOid);

		/* db_schema:{remove_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_tuple:{delete} */
		sepgsql_ts_config_common(confOid, SEPG_DB_TUPLE__DELETE, true);
	}
#endif
}

void
sepgsql_ts_config_comment(Oid confOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{update} */
		sepgsql_ts_config_common(confOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

Oid
sepgsql_ts_dict_create(const char *dictName, Oid namespaceId)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid
			= sepgsql_get_default_tuple_secid(TSDictionaryRelationId);

		/* db_schema:{add_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__ADD_NAME, true);

		/* db_tuple:{insert} */
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_TUPLE,
							 SEPG_DB_TUPLE__INSERT,
							 dictName,
							 true);
		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_ts_dict_alter(Oid dictOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{update} */
		sepgsql_ts_dict_common(dictOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_ts_dict_alter_rename(Oid dictOid, const char *newName)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		namespaceId = get_ts_dict_namespace(dictOid);

		/* db_schema:{add_name remove_name} */
		sepgsql_schema_common(namespaceId,
							  SEPG_DB_SCHEMA__ADD_NAME |
							  SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_tuple:{update} */
		sepgsql_ts_dict_common(dictOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_ts_dict_drop(Oid dictOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		namespaceId = get_ts_dict_namespace(dictOid);

		/* db_schema:{remove_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_tuple:{delete} */
		sepgsql_ts_dict_common(dictOid, SEPG_DB_TUPLE__DELETE, true);
	}
#endif
}

void
sepgsql_ts_dict_comment(Oid dictOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{update} */
        sepgsql_ts_dict_common(dictOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

Oid
sepgsql_ts_parser_create(const char *parseName, Oid namespaceId,
						 Oid startFunc, Oid tokenFunc, Oid endFunc,
						 Oid headlineFunc, Oid lextypeFunc)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid
			= sepgsql_get_default_tuple_secid(TSParserRelationId);

		/* db_schema:{add_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__ADD_NAME, true);

		/* db_procedure:{install} */
		if (OidIsValid(startFunc))
			sepgsql_proc_common(startFunc, SEPG_DB_PROCEDURE__INSTALL, true);
		if (OidIsValid(tokenFunc))
			sepgsql_proc_common(tokenFunc, SEPG_DB_PROCEDURE__INSTALL, true);
		if (OidIsValid(endFunc))
			sepgsql_proc_common(endFunc, SEPG_DB_PROCEDURE__INSTALL, true);
		if (OidIsValid(headlineFunc))
			sepgsql_proc_common(headlineFunc, SEPG_DB_PROCEDURE__INSTALL, true);
		if (OidIsValid(lextypeFunc))
			sepgsql_proc_common(lextypeFunc, SEPG_DB_PROCEDURE__INSTALL, true);

		/* db_tuple:{insert} */
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_TUPLE,
							 SEPG_DB_TUPLE__INSERT,
							 parseName,
							 true);
		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_ts_parser_alter_rename(Oid parseOid, const char *newName)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		namespaceId = get_ts_parser_namespace(parseOid);

		/* db_schema:{add_name remove_name} */
		sepgsql_schema_common(namespaceId,
							  SEPG_DB_SCHEMA__ADD_NAME |
							  SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_tuple:{update} */
		sepgsql_ts_parser_common(parseOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_ts_parser_drop(Oid parseOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		namespaceId = get_ts_parser_namespace(parseOid);

		/* db_schema:{remove_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_tuple:{delete} */
		sepgsql_ts_parser_common(parseOid, SEPG_DB_TUPLE__DELETE, true);
	}
#endif
}

void
sepgsql_ts_parser_comment(Oid parseOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{update} */
		sepgsql_ts_parser_common(parseOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

Oid
sepgsql_ts_template_create(const char *templateName, Oid namespaceId,
						   Oid initFunc, Oid lexizeFunc)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid
			= sepgsql_get_default_tuple_secid(TSTemplateRelationId);

		/* db_schema:{add_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__ADD_NAME, true);

		/* db_procedure:{install} */
		if (OidIsValid(initFunc))
			sepgsql_proc_common(initFunc, SEPG_DB_PROCEDURE__INSTALL, true);
		if (OidIsValid(lexizeFunc))
			sepgsql_proc_common(lexizeFunc, SEPG_DB_PROCEDURE__INSTALL, true);

		/* db_tuple:{insert} */
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_TUPLE,
							 SEPG_DB_TUPLE__INSERT,
							 templateName,
							 true);
		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_ts_template_alter_rename(Oid templateOid, const char *newName)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		namespaceId = get_ts_template_namespace(templateOid);

		/* db_schema:{add_name remove_name} */
		sepgsql_schema_common(namespaceId,
							  SEPG_DB_SCHEMA__ADD_NAME |
							  SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_tuple:{update} */
		sepgsql_ts_template_common(templateOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_ts_template_drop(Oid templateOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid		namespaceId = get_ts_template_namespace(templateOid);

		/* db_schema:{remove_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_tuple:{delete} */
		sepgsql_ts_template_common(templateOid, SEPG_DB_TUPLE__DELETE, true);
	}
#endif
}

void
sepgsql_ts_template_comment(Oid templateOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{update} */
		sepgsql_ts_template_common(templateOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}
