/*
 * type.c
 *
 * SELinux hooks related to types
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "catalog/pg_cast.h"
#include "catalog/pg_seclabel.h"
#include "catalog/pg_type.h"
#include "sepgsql/hooks.h"
#include "sepgsql/sepgsql.h"
#include "utils/builtins.h"
#include "utils/lsyscache.h"
#include "utils/syscache.h"

#ifdef HAVE_SELINUX
bool
sepgsql_type_common(Oid typeOid, uint32 required, bool abort)
{
	sepgsql_sid_t	tsid;
	char		   *auname;
	bool			retval;

	tsid.relid = TypeRelationId;
	tsid.secid = GetSysCacheSecid1(TYPEOID,
								   ObjectIdGetDatum(typeOid));

	auname = format_type_be(typeOid);

	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_TUPLE,
								  required,
								  auname,
								  abort);
	pfree(auname);

	return retval;
}

bool
sepgsql_cast_common(Oid srcTypeOid, Oid dstTypeOid,
					uint32 required, bool abort)
{
	HeapTuple		tuple;
	sepgsql_sid_t	tsid;
	bool			retval;

	tuple = SearchSysCache2(CASTSOURCETARGET,
                            ObjectIdGetDatum(srcTypeOid),
                            ObjectIdGetDatum(dstTypeOid));
    if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for cast (%u,%u)",
			 srcTypeOid, dstTypeOid);

	tsid.relid = CastRelationId;
	tsid.secid = HeapTupleGetSecid(tuple);

	retval = sepgsql_client_perms(tsid,
								  SEPG_CLASS_DB_TUPLE,
								  required,
								  NULL,
								  abort);
	ReleaseSysCache(tuple);

	return retval;
}

static Oid
get_type_namespace(Oid typeOid)
{
	HeapTuple	tuple;
	Oid			namespaceId;

	tuple = SearchSysCache1(TYPEOID, ObjectIdGetDatum(typeOid));
	if (!HeapTupleIsValid(tuple))
		return InvalidOid;

	namespaceId = ((Form_pg_type) GETSTRUCT(tuple))->typnamespace;

	ReleaseSysCache(tuple);

	return namespaceId;
}
#endif

Oid
sepgsql_type_create(const char *typeName, Oid replaced,
					Oid namespaceId, char typeType,
					Oid inputFunc, Oid outputFunc,
					Oid recvFunc, Oid sendFunc,
					Oid modinFunc, Oid modoutFunc, Oid analyzeFunc)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;
		uint32			required;

		/* db_schema:{add_name} */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__ADD_NAME, true);

		/* db_tuple:{insert or update} */
		if (!OidIsValid(replaced))
		{
			nsid = sepgsql_get_default_tuple_secid(TypeRelationId);
			required = SEPG_DB_TUPLE__INSERT;
		}
		else
		{
			nsid.relid = TypeRelationId;
			nsid.secid = GetSysCacheSecid1(TYPEOID,
										   ObjectIdGetDatum(replaced));
			required = SEPG_DB_TUPLE__UPDATE;
		}

		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_TUPLE,
							 required,
							 typeName,
							 true);

		/* db_procedure:{install} */
		if (OidIsValid(inputFunc))
			sepgsql_proc_common(inputFunc, SEPG_DB_PROCEDURE__INSTALL, true);

		/* db_procedure:{install} */
		if (OidIsValid(outputFunc))
			sepgsql_proc_common(outputFunc, SEPG_DB_PROCEDURE__INSTALL, true);

		/* db_procedure:{install} */
		if (OidIsValid(recvFunc))
			sepgsql_proc_common(recvFunc, SEPG_DB_PROCEDURE__INSTALL, true);

		/* db_procedure:{install} */
		if (OidIsValid(sendFunc))
			sepgsql_proc_common(sendFunc, SEPG_DB_PROCEDURE__INSTALL, true);

		/* db_procedure:{install} */
		if (OidIsValid(modinFunc))
			sepgsql_proc_common(modinFunc, SEPG_DB_PROCEDURE__INSTALL, true);

		/* db_procedure:{install} */
		if (OidIsValid(modoutFunc))
			sepgsql_proc_common(modoutFunc, SEPG_DB_PROCEDURE__INSTALL, true);

		/* db_procedure:{install} */
		if (OidIsValid(analyzeFunc))
			sepgsql_proc_common(analyzeFunc, SEPG_DB_PROCEDURE__INSTALL, true);

		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_type_alter(Oid typeOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_type_common(typeOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_type_alter_rename(Oid typeOid, const char *newName)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid	namespaceId = get_type_namespace(typeOid);

		/* db_schema:{add_name remove_name} */
		sepgsql_schema_common(namespaceId,
							  SEPG_DB_SCHEMA__ADD_NAME |
							  SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_tuple:{update} */
		sepgsql_type_common(typeOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

void
sepgsql_type_alter_schema(Oid typeOid, Oid newSchema)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		Oid	namespaceId = get_type_namespace(typeOid);

		/* db_schema:{ remove_name } */
		sepgsql_schema_common(namespaceId, SEPG_DB_SCHEMA__REMOVE_NAME, true);

		/* db_schema:{ add_name } */
		sepgsql_schema_common(newSchema, SEPG_DB_SCHEMA__ADD_NAME, true);

		/* db_tuple:{update} */
		sepgsql_type_common(typeOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

Oid
sepgsql_type_relabel(Oid typeOid, char *newLabel)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	tsid;

		tsid.relid = TypeRelationId;
		tsid.secid = seclabelTransInput(tsid.relid, newLabel);

		/* db_tuple:{update relabelfrom} */
		sepgsql_type_common(typeOid,
							SEPG_DB_TUPLE__UPDATE |
							SEPG_DB_TUPLE__RELABELFROM, true);

		/* db_procedure:{relabelto} */
		sepgsql_client_perms(tsid,
							 SEPG_CLASS_DB_TUPLE,
							 SEPG_DB_TUPLE__RELABELTO,
							 format_type_be(typeOid),
							 true);
		return tsid.secid;
	}
#endif
	ereport(ERROR,
			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
			 errmsg("SE-PostgreSQL is not available")));
	return InvalidOid;
}

void
sepgsql_type_drop(Oid typeOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{delete} */
		sepgsql_type_common(typeOid, SEPG_DB_TUPLE__DELETE, true);
	}
#endif
}

void
sepgsql_type_comment(Oid typeOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		/* db_tuple:{update} */
		sepgsql_type_common(typeOid, SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}

Oid
sepgsql_cast_create(Oid sourceTypeOid, Oid targetTypeOid,
					char castMethod, Oid castFuncOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_sid_t	nsid;

		/* db_tuple:{insert} */
		nsid = sepgsql_get_default_tuple_secid(CastRelationId);
		sepgsql_client_perms(nsid,
							 SEPG_CLASS_DB_TUPLE,
							 SEPG_DB_TUPLE__INSERT,
							 NULL,
							 true);
		/* db_procedure:{install} */
		if (OidIsValid(castFuncOid))
			sepgsql_proc_common(castFuncOid,
								SEPG_DB_PROCEDURE__INSTALL, true);

		return nsid.secid;
	}
#endif
	return InvalidOid;
}

void
sepgsql_cast_drop(Oid sourceTypeOid, Oid targetTypeOid, bool cascade)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_cast_common(sourceTypeOid, targetTypeOid,
							SEPG_DB_TUPLE__DELETE, true);
	}
#endif
}

void
sepgsql_cast_comment(Oid sourceTypeOid, Oid targetTypeOid)
{
#ifdef HAVE_SELINUX
	if (sepgsql_is_enabled())
	{
		sepgsql_cast_common(sourceTypeOid, targetTypeOid,
							SEPG_DB_TUPLE__UPDATE, true);
	}
#endif
}
